# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module codeparser

Analyses the source code to get the structure of a module/script.
This can be used for fictive introspection, and to display the
structure of a source file in for example a tree widget.

"""

# TODO: replace this module, get data from the syntax highlighter in the code editor

import time, threading, re
import pyzo


# Define regular expression patterns
classPattern = r"^\s*"  # Optional whitespace
classPattern += r"(cp?def\s+)?"  # Cython preamble + whitespace
classPattern += r"class\s+"  # The class keyword + whitespace
classPattern += r"([a-zA-Z_][a-zA-Z_0-9]*)\s*"  # The NAME + optional whitespace
classPattern += r"(\(.*?\))?"  # The superclass(es)
classPattern += r"\s*:"  # Optional whitespace and the colon
#
defPattern = r"^\s*"  # Optional whitespace
defPattern += r"(async )?"  # Optional async keyword
defPattern += r"(cp?)?def\s+"  # The Cython preamble, def keyword and whitespace
defPattern += r"([a-zA-Z_][\*a-zA-Z_0-9]*\s+)?"  # Optional Cython return type
defPattern += r"([a-zA-Z_][a-zA-Z_0-9]*)\s*"  # The NAME + optional whitespace
defPattern += r"\((.*?)\)"  # The SIGNATURE
# defPattern += r'\s*:' # Optional whitespace and the colon
# Leave the colon, easier for cython


class Job:
    """Simple class to represent a job."""

    def __init__(self, text, editorId):
        self.text = text
        self.editorId = editorId


class Result:
    """Simple class to represent a parser result."""

    def __init__(self, rootItem, importList, editorId):
        self.rootItem = rootItem
        self.importList = importList
        self.editorId = editorId

    def isMatch(self, editorId):
        """isMatch(editorId):
        Returns whether the result matches with the given editorId.
        The editorId can also be an editor instance."""
        if isinstance(editorId, int):
            return self.editorId == editorId
        else:
            return self.editorId == id(editorId)


class Parser(threading.Thread):
    """Parser
    Parsing sourcecode in a separate thread, this class obtains
    introspection informarion. This class is also the interface
    to the parsed information; it has methods that can be used
    to extract information from the result.
    """

    def __init__(self):
        threading.Thread.__init__(self)

        # Reference current job
        self._job = None

        # Reference to last result
        self._result = None

        # Lock to enable save threading
        self._lock = threading.RLock()

        # Set deamon
        self.daemon = True
        self._exit = False

    def stop(self, timeout=1.0):
        self._exit = True
        self.join(timeout)

    def parseThis(self, editor):
        """parseThis(editor)
        Give the parser new text to parse.
        If the parser is busy parsing text, it will stop doing that
        and start anew with the most recent version of the text.
        """

        # Get text
        text = editor.toPlainText()

        # Make job
        self._lock.acquire()
        self._job = Job(text, id(editor))
        self._lock.release()

    def getFictiveNameSpace(self, editor):
        """getFictiveNameSpace(editor)
        Produce the fictive namespace, based on the current position.
        A list of names is returned.
        """

        # Obtain result
        result = self._getResult()
        if result is None or not result.isMatch(editor):
            return []

        # Get linenr and indent. These are used to establish the namespace
        # based on indentation.
        cursor = editor.textCursor()
        linenr = cursor.blockNumber()
        index = cursor.positionInBlock()

        # init empty namespace and item list
        namespace = []
        items = result.rootItem.children
        curIsClass = False  # to not add methods (of classes)

        while items:
            curitem = None
            for item in items:
                # append name
                if not curIsClass and item.type in ["class", "def"]:
                    namespace.append(item.name)
                # check if this is the one only last one remains
                if (
                    item.type in ["class", "def"]
                    and item.linenr <= linenr
                    and item.linenr2 > linenr
                ):
                    curitem = item
            # prepare for next round
            if curitem and curitem.indent < index:
                items = curitem.children
                if curitem.type == "class":
                    curIsClass = True
                else:
                    curIsClass = False
            else:
                items = []

        return namespace

    def getFictiveClass(self, name, editor, handleSelf=False):
        """getFictiveClass(name, editor, handleSelf=False)
        Return the fictive class object of the given name, or None
        if it does not exist. If handleSelf is True, automatically
        handles "self." names.
        """
        return self._getFictiveItem(name, "class", editor, handleSelf)

    def getFictiveSignature(self, name, editor, handleSelf=False):
        """getFictiveSignature(name, editor, handleSelf=False)
        Get the signature of the fictive function or method of the
        given name. Returns None if the given name is not a known
        function or method. If handleSelf is True, automatically
        handles "self." names.
        """
        # Get item being a function
        item = self._getFictiveItem(name, "def", editor, handleSelf)

        # Get item being a class
        if not item:
            item = self._getFictiveItem(name, "class", editor, handleSelf)
            if item:
                for subItem in item.children:
                    if subItem.name == "__init__" and subItem.type == "def":
                        item = subItem
                        break
                else:
                    item = None

        # Process or return None if there was no item
        if item:
            nameParts = name.split(".")
            return "{}({})".format(nameParts[-1], item.sig)
        else:
            return None

    def getFictiveImports(self, editor):
        """getFictiveImports(editor)
        Get the fictive imports of this source file.
        tuple:
        - list of names that are imported,
        - a dict with the line to import each name
        """

        # Obtain result
        result = self._getResult()
        if result is None or not result.isMatch(editor):
            return [], []

        # Extract list of names and dict of lines
        imports = []
        importlines = {}
        for item in result.importList:
            imports.append(item.name)
            importlines[item.name] = item.text
        return imports, importlines

    def _getResult(self):
        """getResult()
        Savely Obtain result.
        """
        self._lock.acquire()
        result = self._result
        self._lock.release()
        return result

    def _getFictiveItem(self, name, type, editor, handleSelf=False):
        """_getFictiveItem(name, type, editor, handleSelf=False)
        Obtain the fictive item of the given name and type.
        If handleSelf is True, will handle "self." correctly.
        Intended for internal use.
        """

        # Obtain result
        result = self._getResult()
        if result is None or not result.isMatch(editor):
            return None

        # Split name in parts
        nameParts = name.split(".")

        # Try if the first part represents a class instance
        if handleSelf:
            item = self._getFictiveCurrentClass(editor, nameParts[0])
            if item:
                nameParts[0] = item.name

        # Init
        name = nameParts.pop(0)
        items = result.rootItem.children
        theItem = None

        # Search for name
        while items:
            for item in items:
                if item.name == name:
                    # Found it
                    if nameParts:
                        # Go down one level
                        name = nameParts.pop(0)
                        items = item.children
                        break
                    else:
                        # This is it, is it what we wanted?
                        if item.type == type:
                            theItem = item
                            items = []
                            break
            else:
                # Did not find it
                items = []

        return theItem

    def _getFictiveCurrentClass(self, editor, selfname):
        """_getFictiveCurrentClass(editor, selfname)
        Get the fictive object for the class referenced
        using selfname (usually 'self').
        Intendef for internal use.
        """

        # Obtain result
        result = self._getResult()
        if result is None:
            return None

        # Get linenr and indent
        cursor = editor.textCursor()
        linenr = cursor.blockNumber()
        index = cursor.positionInBlock()

        # Init
        items = result.rootItem.children
        theclass = None

        while items:
            curitem = None
            for item in items:
                # check if this is the one only last one remains
                if item.linenr <= linenr:
                    if not item.linenr2 > linenr:
                        continue
                    curitem = item
                    if item.type == "def" and item.selfname == selfname:
                        theclass = item.parent
                else:
                    break

            # prepare for next round
            if curitem and curitem.indent < index:
                items = curitem.children
            else:
                items = []

        # return
        return theclass

    def run(self):
        """run()
        This is the main loop.
        """

        time.sleep(0.5)
        try:
            while True:
                time.sleep(0.1)

                if self._exit:
                    return

                if self._job:

                    # Savely obtain job
                    self._lock.acquire()
                    job = self._job
                    self._job = None
                    self._lock.release()

                    # Analyse job
                    result = self._analyze(job)

                    # Savely store result
                    self._lock.acquire()
                    self._result = result
                    self._lock.release()

                    # Notify
                    if pyzo.editors is not None:
                        pyzo.editors.parserDone.emit()

        except AttributeError:
            pass  # when python exits, time can be None...

    def _analyze(self, job):
        """The core function.
        Analyses the source code.
        Produces:
        - a tree of FictiveObject objects.
        - a (flat) list of the same object
        - a list of imports
        """

        # Remove multiline strings
        text = washMultilineStrings(job.text)

        # Split text in lines
        lines = text.splitlines()
        lines.insert(0, "")  # so the lines start at 1

        # The structure object. It will first only consist of class and defs
        # the rest will be inserted afterwards.
        root = FictiveObject("root", 0, -1, "root")

        # Also keep a flat list (while running this function)
        flatList = []

        # Cells and imports are inserted in the structure afterwards
        leafs = []

        # Keep a list of imports
        importList = []

        # To know when to make something new when for instance a class is defined
        # in an if statement, we keep track of the last valid node/object:
        # Put inside a list, so we can set it from inside a subfuncion
        lastObject = [root]

        # Define funcion to put an item in the structure in the right parent
        def appendToStructure(object):
            # find position in structure to insert
            node = lastObject[0]
            while (object.indent <= node.indent) and (node is not root):
                node = node.parent
            # insert object
            flatList.append(object)
            node.children.append(object)
            object.parent = node
            lastObject[0] = object

        # Find objects!
        # type can be: cell, class, def, import, var
        for i in range(len(lines)):

            # Obtain line
            line = lines[i]

            # Should we stop?
            if self._job or self._exit:
                break

            # Remove indentation
            tmp = line.lstrip()
            indent = len(line) - len(tmp)
            line = tmp.rstrip()

            # Detect cells
            if (
                line.startswith("##")
                or line.startswith("#%%")
                or line.startswith("# %%")
            ):
                if line.startswith("##"):
                    name = line[2:].lstrip()
                elif line.startswith("#%%"):
                    name = line[3:].lstrip()
                else:
                    name = line[4:].lstrip()
                item = FictiveObject("cell", i, indent, name)
                leafs.append(item)
                # Next! (we have to put this before the elif stuff below
                # because it looks like a comment!)
                continue

            # Split in line and comment
            line, tmp, cmnt = line.partition("#")
            line, cmnt = line.rstrip(), cmnt.lower().strip()

            # Detect todos
            if cmnt and (cmnt.startswith("todo:") or cmnt.startswith("2do:")):
                item = FictiveObject("todo", i, indent, cmnt)
                item.linenr2 = i + 1  # a todo is active at one line only
                leafs.append(item)

            # Continue of no line left
            if not line:
                continue

            # Find last valid node. As the indent of the root is set to -1,
            # this will always stop at the root
            while indent <= lastObject[0].indent:
                lastObject[0].linenr2 = i  # close object
                lastObject[0] = lastObject[0].parent

            # Make a lowercase version of the line
            foundSomething = False

            # Detect classes
            if not foundSomething:
                classResult = re.search(classPattern, line)

                if classResult:
                    foundSomething = True
                    # Get name
                    name = classResult.group(2)
                    item = FictiveObject("class", i, indent, name)
                    appendToStructure(item)
                    item.supers = []
                    item.members = []
                    # Get inheritance
                    supers = classResult.group(3)
                    if supers:
                        supers = supers[1:-1].split(",")
                        supers = [tmp.strip() for tmp in supers]
                        item.supers = [tmp for tmp in supers if tmp]

            # Detect functions and methods (also multiline)
            if (not foundSomething) and line.count("def "):
                # Get a multiline version (for long defs)
                multiLine = line
                for ii in range(1, 16):
                    if i + ii < len(lines):
                        multiLine += " " + lines[i + ii].strip()
                # Get result
                defResult = re.search(defPattern, multiLine)
                if defResult:
                    # Get name
                    name = defResult.group(4)
                    item = FictiveObject("def", i, indent, name)
                    appendToStructure(item)
                    item.selfname = None  # will be filled in if a valid method
                    item.sig = defResult.group(5)
                    # is it a method? -> add method to attr and find selfname
                    if item.parent.type == "class":
                        item.parent.members.append(name)

                        # Find what is used as "self"
                        i2 = line.find("(")
                        i4 = line.find(",", i2)
                        if i4 < 0:
                            i4 = line.find(")", i2)
                        if i4 < 0:
                            i4 = i2
                        selfname = line[i2 + 1 : i4].strip()
                        if selfname:
                            item.selfname = selfname

            elif line.count("import "):
                if line.startswith("import "):
                    for name in ParseImport(line[7:]):
                        item = FictiveObject("import", i, indent, name)
                        item.text = line
                        item.linenr2 = i + 1  # an import is active at one line only
                        leafs.append(item)
                        importList.append(item)

                elif line.startswith("from "):
                    i1 = line.find(" import ")
                    for name in ParseImport(line[i1 + 8 :]):
                        if not IsValidName(name):
                            continue  # we cannot do that!
                        item = FictiveObject("import", i, indent, name)
                        item.text = line
                        item.linenr2 = i + 1  # an import is active at one line only
                        leafs.append(item)
                        importList.append(item)

            elif (
                not indent and line.startswith("if __name__ ==") and "__main__" in line
            ):
                item = FictiveObject("nameismain", i, indent, "__main__")
                item.text = line
                appendToStructure(item)

            elif line.count("="):
                if lastObject[0].type == "def" and lastObject[0].selfname:
                    selfname = lastObject[0].selfname + "."
                    line = line.partition("=")[0]
                    if line.count(selfname):
                        # A lot of ifs here. If we got here, the line is part of
                        # a valid method and contains the selfname before the =.
                        # Now we need to establish whether there is a valid
                        # assignment done here...
                        parts = line.split(",")  # handle tuples
                        for part in parts:
                            part = part.strip()
                            part2 = part[len(selfname) :]
                            if part.startswith(selfname) and IsValidName(part2):
                                # add to the list if not already present
                                defItem = lastObject[0]
                                classItem = lastObject[0].parent
                                #
                                item = FictiveObject("attribute", i, indent, part2)
                                item.parent = defItem
                                defItem.children.append(item)
                                if part2 not in classItem.members:
                                    classItem.members.append(part2)

        ## Post processing

        def getTwoItems(series, linenr):
            """Return the two items just above and below the
            given linenr. The object always is a class or def.
            """
            # find object after linenr
            object1, object2 = None, None  # if no items at all
            i = -1
            for i in range(len(series)):
                object = series[i]
                if object.type not in ["class", "def"]:
                    continue
                if object.linenr > linenr:
                    object2 = object
                    break
            # find object just before linenr
            for ii in range(i, -1, -1):
                object = series[ii]
                if object.type not in ["class", "def"]:
                    continue
                if object.linenr < linenr:
                    object1 = object
                    break
            # return result
            return object1, object2

        # insert the leafs (backwards as the last inserted is at the top)
        for leaf in reversed(leafs):
            ob1, ob2 = getTwoItems(flatList, leaf.linenr)
            if ob1 is None:  # also if ob2 is None
                # insert in root
                root.children.insert(0, leaf)
                leaf.parent = root
                continue
            if ob2 is None:
                ob2parent = root
            else:
                ob2parent = ob2.parent

            # get the object IN which to insert it: ob1
            sibling = None
            while 1:
                canGoDeeper = ob1 is not ob2parent
                canGoDeeper = canGoDeeper and ob1 is not root
                shouldGoDeeper = ob1.indent >= leaf.indent
                shouldGoDeeper = shouldGoDeeper or ob1.linenr2 < leaf.linenr
                if canGoDeeper and shouldGoDeeper:
                    sibling = ob1
                    ob1 = ob1.parent
                else:
                    break

            # insert into ob1, after sibling (if available)
            L = ob1.children
            if sibling:
                i = L.index(sibling)
                L.insert(i + 1, leaf)
            else:
                L.insert(0, leaf)

        # Return result
        return Result(root, importList, job.editorId)


## Helper classes and functions


class FictiveObject:
    """An un-instantiated object.
    type can be class, def, import, cell, todo
    extra stuff:
    class   - supers, members
    def     - selfname
    imports - text
    cell    -
    todo    -
    attribute -
    """

    def __init__(self, type, linenr, indent, name):
        self.children = []
        self.type = type
        self.linenr = linenr  # at which line this object starts
        self.linenr2 = 9999999  # at which line it ends
        self.indent = indent
        self.name = name
        self.sig = ""  # for functions and methods


namechars = "abcdefghijklmnopqrstuvwxyz_0123456789"


def IsValidName(name):
    """Given a string, checks whether it is a
    valid name (dots are not valid!)
    """
    if not name:
        return False
    name = name.lower()
    if name[0] not in namechars[0:-10]:
        return False
    tmp = map(lambda x: x not in namechars, name[2:])
    return sum(tmp) == 0


def ParseImport(names):
    for part in names.split(","):
        i1 = part.find(" as ")
        if i1 > 0:
            name = part[i1 + 3 :].strip()
        else:
            name = part.strip()
        yield name


def findString(text, s, i):
    """findString(text, s)
    Find s in text, but only if s is not in a string or commented
    Helper function for washMultilineStrings"""

    while True:
        i = _findString(text, s, i)
        if i < -1:
            i = -i + 1
        else:
            break
    return i


def _findString(text, s, i):
    """Helper function of findString, which is called recursively
    until a match is found, or it is clear there is no match."""

    # Find occurrence
    i2 = text.find(s, i)
    if i2 < 0:
        return -1

    # Find newline  (if none, we're done)
    i1 = text.rfind("\n", 0, i2)
    if i1 < 0:
        return i2

    # Extract the part on the line up to the match
    line = text[i1:i2]

    # Count quotes, we're done if we found none
    if not line.count('"') and not line.count("'") and not line.count("#"):
        return i2

    # So we found quotes, now really count them ...
    prev = ""
    inString = ""  # this is a boolean combined with a flag which quote was used
    isComment = False
    for c in line:
        if c == "#":
            if not inString:
                isComment = True
                break
        elif c in "\"'":
            if not inString:
                inString = c
            elif prev != "\\":
                if inString == c:
                    inString = ""  # exit string
                else:
                    pass  # the other quote can savely be used inside this string
        prev = c

    # If we are in a string, this match is false ...
    if inString or isComment:
        return -i2  # indicate failure and where to continue
    else:
        return i2  # all's right


def washMultilineStrings(text):
    """washMultilineStrings(text)
    Replace all text within multiline strings with dummy chars
    so that it is not parsed.
    """
    i = 0
    s1 = "'''"
    s2 = '"""'
    while i < len(text):
        # Detect start of a multiline comment (there are two versions)
        i1 = findString(text, s1, i)
        i2 = findString(text, s2, i)
        # Stop if nothing found ...
        if i1 == -1 and i2 == -1:
            break
        else:
            # Make no result be very large
            if i1 == -1:
                i1 = 2**60
            if i2 == -1:
                i2 = 2**60
            # Find end of the multiline comment
            if i1 < i2:
                i3 = i1 + 3
                i4 = text.find(s1, i3)
            else:
                i3 = i2 + 3
                i4 = text.find(s2, i3)
            # No end found -> take all text, unclosed string!
            if i4 == -1:
                i4 = 2**32
            # Leave only the first two quotes of the start of the comment
            i3 -= 1
            i4 += 3
            # Replace all non-newline chars
            tmp = re.sub(r"\S", " ", text[i3:i4])
            text = text[:i3] + tmp + text[i3 + len(tmp) :]
            # Prepare for next round
            i = i4 + 1
    return text


"""
## testing skipping of multiline strings
def ThisShouldNotBeVisible():
  pass
class ThisShouldNotBeVisibleEither():
  pass
"""



## Introduction
"""
Welcome to the tutorial for Pyzo! This tutorial should get you
familiarized with Pyzo in just a few minutes. If you feel this tutorial
contains errors or lacks some information, please let us know via
pyzo@googlegroups.com.

Pyzo is a cross-platform Python IDE focused on interactivity and
introspection, which makes it very suitable for scientific computing.
Its practical design is aimed at simplicity and efficiency.

Pyzo consists of two main components, the editor and the shell, and
uses a set of pluggable tools to help the programmer in various ways.

"""


## The editor
"""
The editor (this window) is where your code is located; it is the central
component of Pyzo.

In the editor, each open file is represented as a tab. By right-clicking on
a tab, files can be run, saved, closed, etc.

The right mouse button also enables one to make a file the MAIN FILE of
a project. This file can be recognized by its star symbol and its blue filename,
and it enables running the file more easily (as we will see later in this
tutorial).

For larger projects, the Project manager tool can be used to manage your files
(also described later in this tutorial)
"""


## The shells
"""
The other main component is the window that holds the shells. When Pyzo
starts, a default shell is created. You can add more shells that run
simultaneously, and which may be of different Python versions.

It is good to know that the shells run in a sub-process, such that
when it is busy, Pyzo itself stays responsive, which allows you to
keep coding and even run code in another shell.

Another notable feature is that Pyzo can integrate the event loop of
five different GUI toolkits, thus enabling interactive plotting with
e.g., Visvis or Matplotlib.
  
Via "Shell > Edit shell configurations", you can edit and add shell
configurations. This allows you to for example select the initial
directory, or use a custom PYTHONPATH.

"""


## The tools
"""
Via the "Tools" menu, one can select which tools to use. The tools can
be positioned in any way you want, and can also be un-docked.

Try the "Source Structure" tool to see the outline of this tutorial!

Note that the tools system is designed such that it's quite easy to
create your own tools. Look at the online wiki for more information,
or use one of the existing tools as an example. Also, Pyzo does not
need to restart to see new tools, or to update existing tools.

"""


## Running code
"""
Pyzo supports several ways to run source code in the editor. (see
also the "Run" menu).

  * Run selected lines. If a line is partially selected, the whole
    line is executed. If there is no selection, Pyzo will run the
    current line.
    
  * Run cell. A cell is everything between two commands starting
    with '##', such as the headings in this tutorial. Try running
    the code at the bottom of this cell!

  * Run file. This runs all the code in the current file.
  
  * Run project main file. Runs the code in the current project's
    main file.

Additionally, you can run the current file or the current project's
main file as a script. This will first restart the shell to provide
a clean environment. The shell is also initialized differently:

Things done on shell startup in INTERACTIVE MODE:
  * sys.argv = ['']
  * sys.path is prepended with an empty string (current working directory)
  * The working dir is set to the "Initial directory" of the shell config.
  * The PYTHONSTARTUP script is run.

Things done on shell startup in SCRIPT MODE:
  * __file__ = <script_filename>
  * sys.argv = [ <script_filename> ]
  * sys.path is prepended with the directory containing the script.
  * The working dir is set to the directory containing the script.

Depending on the settings of the Project mananger, the current project
directory may also be inserted in sys.path.
"""

a = 3
b = 4
print("The answer is " + str(a + b))


## The menu
"""
Almost all functionality of Pyzo can be accessed via the menu. For more
advanced/specific stuff, you can use the logger tool (see also
Settings > Advanced settings)

All actions in the menu can be accessed via a shortcut. Change the
shortcuts using the shortcut editor: Settings > Edit key mappings.
  
"""


## Introspection
"""
Pyzo has strong introspection capabilities. Pyzo knows about the objects
in the shell, and parses (not runs) the source code in order to detect
the structure of your code. This enables powerful instospection such
as autocompletion, calltips, interactive help and source structure.
  
"""


## Debugging
"""
Pyzo supports post-mortem debugging, which means that after something
went wrong, you can inspect the stack trace to find the error.

The easiest way to start debugging is to press the "Debug" button
at the upper right corner of the shells.

Once in debug mode, the button becomes expandable, allowing you to
see the stack trace and go to any frame you like. (Starting debug mode
brings you to the bottom frame.) Changing a frame will make all objects
in that frame available in the shell. If possible, Pyzo will also show
the source file belonging to that frame, and select the line where the
error occurred.

Debugging can also be controlled via magic commands, enter "?" in the
shell for more information.

Below follows an example that you can run to test the debugging.
  
"""

import random

someModuleVariable = True


def getNumber():
    return random.choice(range(10))


def foo():
    spam = "yum"
    eggs = 7
    value = bar()


def bar():
    total = 0
    for i1 in range(100):
        i2 = getNumber()
        total += i1 / i2
    return total


foo()

## The Project manager
"""
For working on projects, the Project manager tool can help you to keep
an overview of all your files. To open up the Project manager tool,
select it from the menu (Tools > Project manager).

To add, remove, or edit your projects, click the button with the
wrench icon. In the dialog, select 'New project' to add a project and
select the directory where your project is located. When the project
is added, you can change the Project description (name).

You can select wether the project path is added to the Python
sys.path. This feature allows you to import project modules from the
shell, or from scripts which are not in the project root directory.
Note that this feature requires a restart of the shell to take effect
(Shell > Restart)

The Project manager allows you to switch between your projects easily
using the selection box at the top. The tree view shows the files and
directories in your project. Files can be hidden using the filter that
is specified at the bottom of the Project manager, e.g. !*.pyc to hide
all files that have the extension pyc.
"""



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.


""" pyzokernel/start.py

Starting script for remote processes in pyzo.
This script connects to the pyzo ide using the yoton interface
and imports remote2 to start the interpreter and introspection thread.

Channels
--------
There are four groups of channels. The ctrl channels are streams from
the ide to the kernel and/or broker. The strm channels are streams to
the ide. The stat channels are status channels. The reqp channels are
req/rep channels. All channels are TEXT except for a few OBJECT channels.

ctrl-command: to give simple commands to the interpreter (ala stdin)
ctrl-code (OBJECT): to let the interpreter execute blocks of code
ctrl-broker: to control the broker (restarting etc)

strm-out: the stdout of the interpreter
strm-err: the stderr of the interpreter
strm-raw: the C-level stdout and stderr of the interpreter (captured by broker)
strm-echo: the interpreters echos commands here
strm-prompt: to send the prompts explicitly
strm-broker: for the broker to send messages to the ide
strm-action: for the kernel to push actions to the ide

stat-interpreter): status of the interpreter (ready, busy, very busy, more, etc)
stat-debug (OBJECT): debug status
stat-startup (OBJECT): Used to pass startup parameters to the kernel

reqp-introspect (OBJECT): To query information from the kernel (and for interruping)

"""

# This file is executed with the active directory one up from this file.

import os
import sys
import time
import yoton
import __main__  # we will run code in the __main__.__dict__ namespace


## Make connection object and get channels

# Create a yoton context. All channels are stored at the context.
ct = yoton.Context()

# Create control channels
ct._ctrl_command = yoton.SubChannel(ct, "ctrl-command")
ct._ctrl_code = yoton.SubChannel(ct, "ctrl-code", yoton.OBJECT)

# Create stream channels
ct._strm_out = yoton.PubChannel(ct, "strm-out")
ct._strm_err = yoton.PubChannel(ct, "strm-err")
ct._strm_echo = yoton.PubChannel(ct, "strm-echo")
ct._strm_prompt = yoton.PubChannel(ct, "strm-prompt")
ct._strm_action = yoton.PubChannel(ct, "strm-action", yoton.OBJECT)

# Create status channels
ct._stat_interpreter = yoton.StateChannel(ct, "stat-interpreter")
ct._stat_cd = yoton.StateChannel(ct, "stat-cd")
ct._stat_debug = yoton.StateChannel(ct, "stat-debug", yoton.OBJECT)
ct._stat_startup = yoton.StateChannel(ct, "stat-startup", yoton.OBJECT)
ct._stat_breakpoints = yoton.StateChannel(ct, "stat-breakpoints", yoton.OBJECT)

# Connect (port number given as command line argument)
# Important to do this *before* replacing the stdout etc, because if an
# error occurs here, it will be printed in the shell.
port = int(sys.argv[1])
ct.connect("localhost:" + str(port), timeout=1.0)

# Create file objects for stdin, stdout, stderr
sys.stdin = yoton.FileWrapper(ct._ctrl_command, echo=ct._strm_echo, isatty=True)
sys.stdout = yoton.FileWrapper(ct._strm_out, 256, isatty=True)
sys.stderr = yoton.FileWrapper(ct._strm_err, 256, isatty=True)

# Set fileno on both
sys.stdout.fileno = sys.__stdout__.fileno
sys.stderr.fileno = sys.__stderr__.fileno


## Set Excepthook


def pyzo_excepthook(type, value, tb):
    import sys

    def writeErr(err):
        sys.__stderr__.write(str(err) + "\n")
        sys.__stderr__.flush()

    writeErr("Uncaught exception in interpreter:")
    writeErr(value)
    if not isinstance(value, (OverflowError, SyntaxError, ValueError)):
        while tb:
            writeErr(
                "-> line %i of %s."
                % (tb.tb_frame.f_lineno, tb.tb_frame.f_code.co_filename)
            )
            tb = tb.tb_next
    import time

    time.sleep(0.3)  # Give some time for the message to be send


# Uncomment to detect error in the interpreter itself.
# But better not use it by default. For instance errors in qt events
# are also catched by this function. I think that is because it would
# allow you to show such exceptions in an error dialog.
# sys.excepthook = pyzo_excepthook


## Init interpreter and introspector request channel

# Delay import, so we can detect syntax errors using the except hook
from pyzokernel.interpreter import PyzoInterpreter
from pyzokernel.introspection import PyzoIntrospector

# Create interpreter instance and give dict in which to run all code
__pyzo__ = PyzoInterpreter(__main__.__dict__, "<console>")
sys._pyzoInterpreter = __pyzo__

# Store context
__pyzo__.context = ct

# Create introspection req channel (store at interpreter instance)
__pyzo__.introspector = PyzoIntrospector(ct, "reqp-introspect")


## Clean up

# Delete local variables
del yoton, PyzoInterpreter, PyzoIntrospector, pyzo_excepthook
del ct, port
del os, sys, time

# Delete stuff we do not want
for name in [
    "__file__",  # __main__ does not have a corresponding file
    "__loader__",  # prevent lines from this file to be shown in tb
]:
    globals().pop(name, None)
del name


## Start and stop

# Start introspector and enter the interpreter
__pyzo__.introspector.set_mode("thread")

try:
    __pyzo__.run()

finally:
    # Restore original streams, so that SystemExit behaves as intended
    import sys

    try:
        sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__
    except Exception:
        pass
    # Flush pending messages (raises exception if times out)
    try:
        __pyzo__.context.flush(0.1)
    except Exception:
        pass
    # Nicely exit by closing context (closes channels and connections). If we do
    # not do this on Python 3.2 (at least Windows) the exit delays 10s. (issue 79)
    try:
        __pyzo__.introspector.set_mode(0)
        __pyzo__.context.close()
    except Exception:
        pass



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.


import pyzo
from pyzo.qt import QtCore, QtGui, QtWidgets

tool_name = pyzo.translate("pyzoWorkspace", "Workspace")
tool_summary = pyzo.translate(
    "pyzoWorkspace", "Lists the variables in the current shell's namespace."
)


def splitName(name):
    """splitName(name)
    Split an object name in parts, taking dots and indexing into account.
    """
    name = name.replace("[", ".[")
    parts = name.split(".")
    return [p for p in parts if p]


def joinName(parts):
    """joinName(parts)
    Join the parts of an object name, taking dots and indexing into account.
    """
    name = ".".join(parts)
    return name.replace(".[", "[")


class WorkspaceProxy(QtCore.QObject):
    """WorkspaceProxy

    A proxy class to handle the asynchonous behaviour of getting information
    from the shell. The workspace tool asks for a certain name, and this
    class notifies when new data is available using a qt signal.

    """

    haveNewData = QtCore.Signal()

    def __init__(self):
        QtCore.QObject.__init__(self)

        # Variables
        self._variables = []

        # Element to get more info of
        self._name = ""

        # Bind to events
        pyzo.shells.currentShellChanged.connect(self.onCurrentShellChanged)
        pyzo.shells.currentShellStateChanged.connect(self.onCurrentShellStateChanged)

        # Initialize
        self.onCurrentShellStateChanged()

    def addNamePart(self, part):
        """addNamePart(part)
        Add a part to the name.
        """
        parts = splitName(self._name)
        parts.append(part)
        self.setName(joinName(parts))

    def setName(self, name):
        """setName(name)
        Set the name that we want to know more of.
        """
        self._name = name

        shell = pyzo.shells.getCurrentShell()
        if shell:
            future = shell._request.dir2(self._name)
            future.add_done_callback(self.processResponse)

    def goUp(self):
        """goUp()
        Cut the last part off the name.
        """
        parts = splitName(self._name)
        if parts:
            parts.pop()
        self.setName(joinName(parts))

    def onCurrentShellChanged(self):
        """onCurrentShellChanged()
        When no shell is selected now, update this. In all other cases,
        the onCurrentShellStateChange will be fired too.
        """
        shell = pyzo.shells.getCurrentShell()
        if not shell:
            self._variables = []
            self.haveNewData.emit()

    def onCurrentShellStateChanged(self):
        """onCurrentShellStateChanged()
        Do a request for information!
        """
        shell = pyzo.shells.getCurrentShell()
        if not shell:
            # Should never happen I think, but just to be sure
            self._variables = []
        elif shell._state.lower() != "busy":
            future = shell._request.dir2(self._name)
            future.add_done_callback(self.processResponse)

    def processResponse(self, future):
        """processResponse(response)
        We got a response, update our list and notify the tree.
        """

        response = []

        # Process future
        if future.cancelled():
            pass  # print('Introspect cancelled') # No living kernel
        elif future.exception():
            print("Introspect-queryDoc-exception: ", future.exception())
        else:
            response = future.result()

        self._variables = response
        self.haveNewData.emit()


class WorkspaceItem(QtWidgets.QTreeWidgetItem):
    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()
        try:
            return float(self.text(column).strip("[]")) > float(
                otherItem.text(column).strip("[]")
            )
        except ValueError:
            return self.text(column) > otherItem.text(column)


class WorkspaceTree(QtWidgets.QTreeWidget):
    """WorkspaceTree

    The tree that displays the items in the current namespace.
    I first thought about implementing this using the mode/view
    framework, but it is so much work and I can't seem to fully
    understand how it works :(

    The QTreeWidget is so very simple and enables sorting very
    easily, so I'll stick with that ...

    """

    def __init__(self, parent):
        QtWidgets.QTreeWidget.__init__(self, parent)

        self._config = parent._config

        # Set header stuff
        self.setHeaderHidden(False)
        self.setColumnCount(3)
        self.setHeaderLabels(
            [
                pyzo.translate("pyzoWorkspace", "Name"),
                pyzo.translate("pyzoWorkspace", "Type"),
                pyzo.translate("pyzoWorkspace", "Repr"),
            ]
        )
        # self.setColumnWidth(0, 100)
        self.setSortingEnabled(True)

        # Nice rows
        self.setAlternatingRowColors(True)
        self.setRootIsDecorated(False)

        # Create proxy
        self._proxy = WorkspaceProxy()
        self._proxy.haveNewData.connect(self.fillWorkspace)

        # For menu
        self.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)
        self._menu = QtWidgets.QMenu()
        self._menu.triggered.connect(self.contextMenuTriggered)

        # Bind to events
        self.itemActivated.connect(self.onItemExpand)

        self._startUpVariables = ["In", "Out", "exit", "get_ipython", "quit"]

    def contextMenuEvent(self, event):
        """contextMenuEvent(event)
        Show the context menu.
        """

        QtWidgets.QTreeView.contextMenuEvent(self, event)

        # Get if an item is selected
        item = self.currentItem()
        if not item:
            return

        # Create menu
        self._menu.clear()
        commands = [
            ("Show namespace", pyzo.translate("pyzoWorkspace", "Show namespace")),
            ("Show help", pyzo.translate("pyzoWorkspace", "Show help")),
            ("Delete", pyzo.translate("pyzoWorkspace", "Delete")),
        ]
        for a, display in commands:
            action = self._menu.addAction(display)
            action._what = a
            parts = splitName(self._proxy._name)
            parts.append(item.text(0))
            action._objectName = joinName(parts)
            action._item = item

        # Show
        self._menu.popup(QtGui.QCursor.pos() + QtCore.QPoint(3, 3))

    def contextMenuTriggered(self, action):
        """contextMenuTriggered(action)
        Process a request from the context menu.
        """

        # Get text
        req = action._what.lower()

        if "namespace" in req:
            # Go deeper
            self.onItemExpand(action._item)

        elif "help" in req:
            # Show help in help tool (if loaded)
            hw = pyzo.toolManager.getTool("pyzointeractivehelp")
            if hw:
                hw.setObjectName(action._objectName, addToHist=True)

        elif "delete" in req:
            # Delete the variable
            shell = pyzo.shells.getCurrentShell()
            if shell:
                shell.processLine("del " + action._objectName)

    def onItemExpand(self, item):
        """onItemExpand(item)
        Inspect the attributes of that item.
        """
        self._proxy.addNamePart(item.text(0))

    def fillWorkspace(self):
        """fillWorkspace()
        Update the workspace tree.
        """

        # Clear first
        self.clear()

        # Set name
        line = self.parent()._line
        line.setText(self._proxy._name)

        # Add elements
        for des in self._proxy._variables:

            # Get parts
            parts = list(des)
            if len(parts) < 4:
                continue

            name = parts[0]

            # Pop the 'kind' element
            kind = parts.pop(2)

            # <kludge 2>
            # the typeTranslation dictionary contains "synonyms" for types that will be hidden
            # Currently only "method"->"function" is used
            # the try:... is there to have a minimal translation dictionary.
            try:
                kind = self._config.typeTranslation[kind]
            except KeyError:
                pass
            # </kludge 2>
            if kind in self._config.hideTypes:
                continue
            if name.startswith("_") and "private" in self._config.hideTypes:
                continue
            if "startup" in self._config.hideTypes and name in self._startUpVariables:
                continue

            # Create item
            item = WorkspaceItem(parts, 0)
            self.addTopLevelItem(item)

            # Set tooltip
            tt = "%s: %s" % (parts[0], parts[-1])
            item.setToolTip(0, tt)
            item.setToolTip(1, tt)
            item.setToolTip(2, tt)

        self.parent().displayEmptyWorkspace(
            self.topLevelItemCount() == 0 and self._proxy._name == ""
        )


class PyzoWorkspace(QtWidgets.QWidget):
    """PyzoWorkspace

    The main widget for this tool.

    """

    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent)

        # Make sure there is a configuration entry for this tool
        # The pyzo tool manager makes sure that there is an entry in
        # config.tools before the tool is instantiated.
        toolId = self.__class__.__name__.lower()
        self._config = pyzo.config.tools[toolId]
        if not hasattr(self._config, "hideTypes"):
            self._config.hideTypes = []
        # <kludge 2>
        # configuring the typeTranslation dictionary
        if not hasattr(self._config, "typeTranslation"):
            # to prevent the exception to be raised, one could init to :
            # {"method": "function", "function": "function", "type": "type", "private": "private", "module": "module"}
            self._config.typeTranslation = {}
        # Defaults
        self._config.typeTranslation["method"] = "function"
        self._config.typeTranslation["builtin_function_or_method"] = "function"
        # <kludge 2>

        # Create tool button
        self._up = QtWidgets.QToolButton(self)
        style = QtWidgets.qApp.style()
        self._up.setIcon(style.standardIcon(style.SP_ArrowLeft))
        self._up.setIconSize(QtCore.QSize(16, 16))

        # Create "path" line edit
        self._line = QtWidgets.QLineEdit(self)
        self._line.setReadOnly(True)
        self._line.setStyleSheet("QLineEdit { background:#ddd; }")
        self._line.setFocusPolicy(QtCore.Qt.NoFocus)

        # Create options menu
        self._options = QtWidgets.QToolButton(self)
        self._options.setIcon(pyzo.icons.filter)
        self._options.setIconSize(QtCore.QSize(16, 16))
        self._options.setPopupMode(self._options.InstantPopup)
        self._options.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        #
        self._options._menu = QtWidgets.QMenu()
        self._options.setMenu(self._options._menu)
        self.onOptionsPress()  # create menu now

        # Create tree
        self._tree = WorkspaceTree(self)

        # Create message for when tree is empty
        self._initText = QtWidgets.QLabel(
            pyzo.translate(
                "pyzoWorkspace",
                """Lists the variables in the current shell's namespace.

Currently, there are none. Some of them may be hidden because of the filters you configured.""",
            ),
            self,
        )
        self._initText.setVisible(False)
        self._initText.setWordWrap(True)

        # Set layout
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self._up, 0)
        layout.addWidget(self._line, 1)
        layout.addWidget(self._options, 0)
        #
        mainLayout = QtWidgets.QVBoxLayout(self)
        mainLayout.addLayout(layout, 0)
        mainLayout.addWidget(self._initText, 1)
        mainLayout.addWidget(self._tree, 2)
        mainLayout.setSpacing(2)
        # set margins
        margin = pyzo.config.view.widgetMargin
        mainLayout.setContentsMargins(margin, margin, margin, margin)
        self.setLayout(mainLayout)

        # Bind events
        self._up.pressed.connect(self._tree._proxy.goUp)
        self._options.pressed.connect(self.onOptionsPress)
        self._options._menu.triggered.connect(self.onOptionMenuTiggered)

    def displayEmptyWorkspace(self, empty):
        self._tree.setVisible(not empty)
        self._initText.setVisible(empty)

    def onOptionsPress(self):
        """Create the menu for the button, Do each time to make sure
        the checks are right."""

        # Get menu
        menu = self._options._menu
        menu.clear()

        hideables = [
            ("type", pyzo.translate("pyzoWorkspace", "Hide types")),
            ("function", pyzo.translate("pyzoWorkspace", "Hide functions")),
            ("module", pyzo.translate("pyzoWorkspace", "Hide modules")),
            ("private", pyzo.translate("pyzoWorkspace", "Hide private identifiers")),
            (
                "startup",
                pyzo.translate("pyzoWorkspace", "Hide the shell's startup variables"),
            ),
        ]

        for type, display in hideables:
            checked = type in self._config.hideTypes
            action = menu.addAction(display)
            action._what = type
            action.setCheckable(True)
            action.setChecked(checked)

    def onOptionMenuTiggered(self, action):
        """The user decides what to hide in the workspace."""

        # What to show
        type = action._what.lower()

        # Swap
        if type in self._config.hideTypes:
            while type in self._config.hideTypes:
                self._config.hideTypes.remove(type)
        else:
            self._config.hideTypes.append(type)

        # Update
        self._tree.fillWorkspace()



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module main

This module contains the main frame. Implements the main window.
Also adds some variables to the pyzo namespace, such as the callLater
function which is also defined here.

"""

import os, sys, time
import base64
from queue import Queue, Empty

import pyzo
from pyzo.core.icons import IconArtist
from pyzo.core import commandline
from pyzo.core.statusbar import StatusBar
from pyzo import qt
from pyzo.qt import QtCore, QtGui, QtWidgets
from pyzo.core.splash import SplashWidget
from pyzo.util import paths
from pyzo.util import zon as ssdf  # zon is ssdf-light
from pyzo import translate


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, parent=None, locale=None):
        QtWidgets.QMainWindow.__init__(self, parent)

        self._closeflag = 0  # Used during closing/restarting

        # Init window title and application icon
        # Set title to something nice. On Ubuntu 12.10 this text is what
        # is being shown at the fancy title bar (since it's not properly
        # updated)
        self.setMainTitle()
        loadAppIcons()
        self.setWindowIcon(pyzo.icon)

        # Restore window geometry before drawing for the first time,
        # such that the window is in the right place
        self.resize(800, 600)  # default size
        self.restoreGeometry()

        # Show splash screen (we need to set our color too)
        w = SplashWidget(self, distro="no distro")
        self.setCentralWidget(w)
        self.setStyleSheet("QMainWindow { background-color: #268bd2;}")

        # Show empty window and disable updates for a while
        self.show()
        self.paintNow()
        self.setUpdatesEnabled(False)

        # Determine timeout for showing splash screen
        splash_timeout = time.time() + 1.0

        # Set locale of main widget, so that qt strings are translated
        # in the right way
        if locale:
            self.setLocale(locale)

        # Store myself
        pyzo.main = self

        # Init dockwidget settings
        self.setTabPosition(QtCore.Qt.AllDockWidgetAreas, QtWidgets.QTabWidget.South)
        self.setDockOptions(
            QtWidgets.QMainWindow.AllowTabbedDocks
            | QtWidgets.QMainWindow.AllowNestedDocks
            # |  QtWidgets.QMainWindow.AnimatedDocks
        )

        # Set window atrributes
        self.setAttribute(QtCore.Qt.WA_AlwaysShowToolTips, True)

        # Load icons and fonts
        loadIcons()
        loadFonts()

        # Set qt style and test success
        self.setQtStyle(None)  # None means init!

        # Hold the splash screen if needed
        while time.time() < splash_timeout:
            QtWidgets.qApp.sendPostedEvents()
            QtWidgets.qApp.processEvents()
            time.sleep(0.05)

        # Populate the window (imports more code)
        self._populate()

        # Revert to normal background, and enable updates
        self.setStyleSheet("")
        self.setUpdatesEnabled(True)

        # Restore window state, force updating, and restore again
        self.restoreState()
        self.paintNow()
        self.restoreState()

        # Present user with wizard if he/she is new.
        if False:  # pyzo.config.state.newUser:
            from pyzo.util.pyzowizard import PyzoWizard

            w = PyzoWizard(self)
            w.show()  # Use show() instead of exec_() so the user can interact with pyzo

        # Create new shell config if there is None
        if not pyzo.config.shellConfigs2:
            from pyzo.core.kernelbroker import KernelInfo

            pyzo.config.shellConfigs2.append(KernelInfo())

        # Focus on editor
        e = pyzo.editors.getCurrentEditor()
        if e is not None:
            e.setFocus()

        # Handle any actions
        commandline.handle_cmd_args()

    # To force drawing ourselves
    def paintEvent(self, event):
        QtWidgets.QMainWindow.paintEvent(self, event)
        self._ispainted = True

    def paintNow(self):
        """Enforce a repaint and keep calling processEvents until
        we are repainted.
        """
        self._ispainted = False
        self.update()
        while not self._ispainted:
            QtWidgets.qApp.sendPostedEvents()
            QtWidgets.qApp.processEvents()
            time.sleep(0.01)

    def _populate(self):

        # Delayed imports
        from pyzo.core.editorTabs import EditorTabs
        from pyzo.core.shellStack import ShellStackWidget
        from pyzo.core import codeparser
        from pyzo.core.history import CommandHistory
        from pyzo.tools import ToolManager

        # Instantiate tool manager
        pyzo.toolManager = ToolManager()

        # Check to install conda now ...
        # from pyzo.util.bootstrapconda import check_for_conda_env
        # check_for_conda_env()

        # Instantiate and start source-code parser
        if pyzo.parser is None:
            pyzo.parser = codeparser.Parser()
            pyzo.parser.start()

        # Create editor stack and make the central widget
        pyzo.editors = EditorTabs(self)
        self.setCentralWidget(pyzo.editors)

        # Create floater for shell
        self._shellDock = dock = QtWidgets.QDockWidget(self)
        if pyzo.config.settings.allowFloatingShell:
            dock.setFeatures(dock.DockWidgetMovable | dock.DockWidgetFloatable)
        else:
            dock.setFeatures(dock.DockWidgetMovable)
        dock.setObjectName("shells")
        dock.setWindowTitle("Shells")
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)

        # Create shell stack
        pyzo.shells = ShellStackWidget(self)
        dock.setWidget(pyzo.shells)

        # Initialize command history
        pyzo.command_history = CommandHistory("command_history.py")

        # Create the default shell when returning to the event queue
        callLater(pyzo.shells.addShell)

        # Create status bar
        self.setStatusBar(StatusBar())
        # show or hide
        self.statusBar().setVisible(pyzo.config.view.showStatusbar)

        # Create menu
        from pyzo.core import menu

        pyzo.keyMapper = menu.KeyMapper()
        menu.buildMenus(self.menuBar())

        # Add the context menu to the editor
        pyzo.editors.addContextMenu()
        pyzo.shells.addContextMenu()

        # Load tools
        if pyzo.config.state.newUser and not pyzo.config.state.loadedTools:
            pyzo.toolManager.loadTool("pyzosourcestructure")
            pyzo.toolManager.loadTool("pyzofilebrowser", "pyzosourcestructure")
        elif pyzo.config.state.loadedTools:
            for toolId in pyzo.config.state.loadedTools:
                pyzo.toolManager.loadTool(toolId)

    def setMainTitle(self, path=None):
        """Set the title of the main window, by giving a file path."""
        if not path:
            # Plain title
            title = "Interactive Editor for Python"
        else:
            # Title with a filename
            name = os.path.basename(path)
            if os.path.isfile(path):
                pass
            elif name == path:
                path = translate("main", "unsaved")
            else:
                pass  # We hope the given path is informative
            # Set title
            tmp = {
                "fileName": name,
                "filename": name,
                "name": name,
                "fullPath": path,
                "fullpath": path,
                "path": path,
            }
            title = pyzo.config.advanced.titleText.format(**tmp)

        # Set
        self.setWindowTitle(title)

    def saveWindowState(self):
        """Save:
        * which tools are loaded
        * geometry of the top level windows
        * layout of dockwidgets and toolbars
        """

        # Save tool list
        tools = pyzo.toolManager.getLoadedTools()
        pyzo.config.state.loadedTools = tools

        # Store window geometry
        geometry = self.saveGeometry()
        geometry = base64.encodebytes(geometry.data()).decode("ascii")
        pyzo.config.state.windowGeometry = geometry

        # Store window state
        state = self.saveState()
        state = base64.encodebytes(state.data()).decode("ascii")
        pyzo.config.state.windowState = state

    def restoreGeometry(self, value=None):
        # Restore window position and whether it is maximized

        if value is not None:
            return super().restoreGeometry(value)

        # No value give, try to get it from the config
        if pyzo.config.state.windowGeometry:
            try:
                geometry = pyzo.config.state.windowGeometry
                geometry = base64.decodebytes(geometry.encode("ascii"))
                self.restoreGeometry(geometry)
            except Exception as err:
                print("Could not restore window geomerty: " + str(err))

    def restoreState(self, value=None):
        # Restore layout of dock widgets and toolbars

        if value is not None:
            return super().restoreState(value)

        # No value give, try to get it from the config
        if pyzo.config.state.windowState:
            try:
                state = pyzo.config.state.windowState
                state = base64.decodebytes(state.encode("ascii"))
                self.restoreState(state)
            except Exception as err:
                print("Could not restore window state: " + str(err))

    def setQtStyle(self, stylename=None):
        """Set the style and the palette, based on the given style name.
        If stylename is None or not given will do some initialization.
        If bool(stylename) evaluates to False will use the default style
        for this system. Returns the QStyle instance.
        """

        if stylename is None:
            # Initialize

            # Get native pallette (used below)
            QtWidgets.qApp.nativePalette = QtWidgets.qApp.palette()

            # Obtain default style name
            pyzo.defaultQtStyleName = str(QtWidgets.qApp.style().objectName())

            # Other than gtk+ and mac, Fusion/Cleanlooks looks best (in my opinion)
            if "gtk" in pyzo.defaultQtStyleName.lower():
                pass  # Use default style
            elif "macintosh" in pyzo.defaultQtStyleName.lower():
                pass  # Use default style
            elif qt.QT_VERSION > "5":
                pyzo.defaultQtStyleName = "Fusion"
            else:
                pyzo.defaultQtStyleName = "Cleanlooks"

            # Set style if there is no style yet
            if not pyzo.config.view.qtstyle:
                pyzo.config.view.qtstyle = pyzo.defaultQtStyleName

        # Init
        if not stylename:
            stylename = pyzo.config.view.qtstyle

        # Check if this style exist, set to default otherwise
        styleNames = [name.lower() for name in QtWidgets.QStyleFactory.keys()]
        if stylename.lower() not in styleNames:
            stylename = pyzo.defaultQtStyleName

        # Try changing the style
        qstyle = QtWidgets.qApp.setStyle(stylename)

        # Set palette
        if qstyle:
            QtWidgets.qApp.setPalette(QtWidgets.qApp.nativePalette)

        # Done
        return qstyle

    def closeEvent(self, event):
        """Override close event handler."""

        # Are we restaring?
        restarting = time.time() - self._closeflag < 1.0  # noqa: F841

        # Save settings
        pyzo.saveConfig()
        pyzo.command_history.save()

        # Stop command server
        commandline.stop_our_server()

        # Proceed with closing...
        result = pyzo.editors.closeAll()
        if not result:
            self._closeflag = False
            event.ignore()
            return
        else:
            self._closeflag = True
            # event.accept()  # Had to comment on Windows+py3.3 to prevent error

        # Proceed with closing shells
        pyzo.localKernelManager.terminateAll()

        for shell in pyzo.shells:
            shell._context.close()

        # The tools need to be explicitly closed to allow them to clean up
        for toolname in pyzo.toolManager.getLoadedTools():
            tool = pyzo.toolManager.getTool(toolname)
            if hasattr(tool, "cleanUp"):
                tool.cleanUp()

        # Stop all threads (this should really only be daemon threads)
        import threading

        for thread in threading.enumerate():
            if hasattr(thread, "stop"):
                try:
                    thread.stop(0.1)
                except Exception:
                    pass

        #         # Wait for threads to die ...
        #         # This should not be necessary, but I used it in the hope that it
        #         # would prevent the segfault on Python3.3. It didn't.
        #         timeout = time.time() + 0.5
        #         while threading.activeCount() > 1 and time.time() < timeout:
        #             time.sleep(0.1)
        #         print('Number of threads alive:', threading.activeCount())

        # Proceed as normal
        QtWidgets.QMainWindow.closeEvent(self, event)

    def restart(self):
        """Restart Pyzo."""

        self._closeflag = time.time()

        # Close
        self.close()

        if self._closeflag:
            # Get args
            args = [arg for arg in sys.argv]

            if not paths.is_frozen():
                # Prepend the executable name (required on Linux)
                lastBit = os.path.basename(sys.executable)
                args.insert(0, lastBit)

            # When running from the pip entry point pyzo.exe ... (issue #641)
            if (
                len(args) == 2
                and args[0] == "python.exe"
                and not os.path.isfile(args[1])
            ):
                args = ["python.exe", "-m", "pyzo"]

            # Replace the process!
            os.execv(sys.executable, args)

    def createPopupMenu(self):

        # Init menu
        menu = QtWidgets.QMenu()

        # Insert two items
        for item in ["Editors", "Shells"]:
            action = menu.addAction(item)
            action.setCheckable(True)
            action.setChecked(True)
            action.setEnabled(False)

        # Insert tools
        for tool in pyzo.toolManager.loadToolInfo():
            action = menu.addAction(tool.name)
            action.setCheckable(True)
            action.setChecked(bool(tool.instance))
            action.menuLauncher = tool.menuLauncher

        # Show menu and process result
        a = menu.popup(QtGui.QCursor.pos())
        if a:
            a.menuLauncher(not a.menuLauncher(None))


def loadAppIcons():
    """loadAppIcons()
    Load the application iconsr.
    """
    # Get directory containing the icons
    appiconDir = os.path.join(pyzo.pyzoDir, "resources", "appicons")

    # Determine template for filename of the application icon-files.
    fnameT = "pyzologo{}.png"

    # Construct application icon. Include a range of resolutions. Note that
    # Qt somehow does not use the highest possible res on Linux/Gnome(?), even
    # the logo of qt-designer when alt-tabbing looks a bit ugly.
    pyzo.icon = QtGui.QIcon()
    for sze in [16, 32, 48, 64, 128, 256]:
        fname = os.path.join(appiconDir, fnameT.format(sze))
        if os.path.isfile(fname):
            pyzo.icon.addFile(fname, QtCore.QSize(sze, sze))

    # Set as application icon. This one is used as the default for all
    # windows of the application.
    QtWidgets.qApp.setWindowIcon(pyzo.icon)

    # Construct another icon to show when the current shell is busy
    artist = IconArtist(pyzo.icon)  # extracts the 16x16 version
    artist.setPenColor("#0B0")
    for x in range(11, 16):
        d = x - 11  # runs from 0 to 4
        artist.addLine(x, 6 + d, x, 15 - d)
    pm = artist.finish().pixmap(16, 16)
    #
    pyzo.iconRunning = QtGui.QIcon(pyzo.icon)
    pyzo.iconRunning.addPixmap(pm)  # Change only 16x16 icon


def loadIcons():
    """loadIcons()
    Load all icons in the icon dir.
    """
    # Get directory containing the icons
    iconDir = os.path.join(pyzo.pyzoDir, "resources", "icons")

    # Construct other icons
    dummyIcon = IconArtist().finish()
    pyzo.icons = ssdf.new()
    for fname in os.listdir(iconDir):
        if fname.endswith(".png"):
            try:
                # Short and full name
                name = fname.split(".")[0]
                name = name.replace("pyzo_", "")  # discart prefix
                ffname = os.path.join(iconDir, fname)
                # Create icon
                icon = QtGui.QIcon()
                icon.addFile(ffname, QtCore.QSize(16, 16))
                # Store
                pyzo.icons[name] = icon
            except Exception as err:
                pyzo.icons[name] = dummyIcon
                print("Could not load icon %s: %s" % (fname, str(err)))


def loadFonts():
    """loadFonts()
    Load all fonts that come with Pyzo.
    """
    import pyzo.codeeditor  # we need pyzo and codeeditor namespace here

    # Get directory containing the icons
    fontDir = os.path.join(pyzo.pyzoDir, "resources", "fonts")

    # Get database object
    db = QtGui.QFontDatabase  # static class

    # Set default font
    pyzo.codeeditor.Manager.setDefaultFontFamily("DejaVu Sans Mono")

    # Load fonts that are in the fonts directory
    if os.path.isdir(fontDir):
        for fname in os.listdir(fontDir):
            if "oblique" in fname.lower():  # issue #461
                continue
            if os.path.splitext(fname)[1].lower() in [".otf", ".ttf"]:
                try:
                    db.addApplicationFont(os.path.join(fontDir, fname))
                except Exception as err:
                    print("Could not load font %s: %s" % (fname, str(err)))


class _CallbackEventHandler(QtCore.QObject):
    """Helper class to provide the callLater function."""

    def __init__(self):
        QtCore.QObject.__init__(self)
        self.queue = Queue()

    def customEvent(self, event):
        while True:
            try:
                callback, args = self.queue.get_nowait()
            except Empty:
                break
            try:
                callback(*args)
            except Exception as why:
                print("callback failed: {}:\n{}".format(callback, why))

    def postEventWithCallback(self, callback, *args):
        self.queue.put((callback, args))
        QtWidgets.qApp.postEvent(self, QtCore.QEvent(QtCore.QEvent.User))


def callLater(callback, *args):
    """callLater(callback, *args)
    Post a callback to be called in the main thread.
    """
    _callbackEventHandler.postEventWithCallback(callback, *args)


# Create callback event handler instance and insert function in pyzo namespace
_callbackEventHandler = _CallbackEventHandler()
pyzo.callLater = callLater


_SCREENSHOT_CODE = """
import random

numerator = 4

def get_number():
    # todo: something appears to be broken here
    val = random.choice(range(10))
    return numerator / val

class Groceries(list):
    \"\"\" Overloaded list class.
    \"\"\"
    def append_defaults(self):
        spam = 'yum'
        pie = 3.14159
        self.extend([spam, pie])

class GroceriesPlus(Groceries):
    \"\"\" Groceries with surprises!
    \"\"\"
    def append_random(self):
        value = get_number()
        self.append(value)

# Create some groceries
g = GroceriesPlus()
g.append_defaults()
g.append_random()

"""


def screenshotExample(width=1244, height=700):
    e = pyzo.editors.newFile()
    e.editor.setPlainText(_SCREENSHOT_CODE)
    pyzo.main.resize(width, height)


def screenshot(countdown=5):
    QtCore.QTimer.singleShot(countdown * 1000, _screenshot)


def _screenshot():
    # Grab
    print("SNAP!")
    pix = QtGui.QPixmap.grabWindow(pyzo.main.winId())
    # pix = QtGui.QPixmap.grabWidget(pyzo.main)
    # Get name
    i = 1
    while i > 0:
        name = "pyzo_screen_%s_%02i.png" % (sys.platform, i)
        fname = os.path.join(os.path.expanduser("~"), name)
        if os.path.isfile(fname):
            i += 1
        else:
            i = -1
    # Save screenshot and a thumb
    pix.save(fname)
    thumb = pix.scaledToWidth(500, QtCore.Qt.SmoothTransformation)
    thumb.save(fname.replace("screen", "thumb"))
    print("Screenshot and thumb saved in", os.path.expanduser("~"))


pyzo.screenshot = screenshot
pyzo.screenshotExample = screenshotExample



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Icons module

Defines functionality for creating icons by composing different overlays
and also by directly drawing into the pixmap. This allows making icons
that show information to the user in a very effective, yet subtle manner.

"""

from pyzo.qt import QtCore, QtGui, QtWidgets
import pyzo


class IconArtist:
    """IconArtist(icon=None)

    Object to draw icons with. Can be instantiated with an existing icon
    or as a blank icon. Perform operations and then use finish() to
    obtain the result.

    """

    def __init__(self, icon=None):

        # Get pixmap from given icon (None creates empty pixmap)
        self._pm = self._getPixmap(icon)

        # Instantiate painter for the pixmap
        self._painter = QtGui.QPainter()
        self._painter.begin(self._pm)

    def finish(self, icon=None):
        """finish()
        Finish the drawing and return the resulting icon.
        """
        self._painter.end()
        return QtGui.QIcon(self._pm)

    def _getPixmap(self, icon):

        # Get icon if given by name
        if isinstance(icon, str):
            icon = pyzo.icons[icon]

        # Create pixmap
        if icon is None:
            pm = QtGui.QPixmap(16, 16)
            pm.fill(QtGui.QColor(0, 0, 0, 0))
            return pm
        elif isinstance(icon, tuple):
            pm = QtGui.QPixmap(icon[0], icon[1])
            pm.fill(QtGui.QColor(0, 0, 0, 0))
            return pm
        elif isinstance(icon, QtGui.QPixmap):
            return icon
        elif isinstance(icon, QtGui.QIcon):
            return icon.pixmap(16, 16)
        else:
            raise ValueError("Icon for IconArtis should be icon, pixmap or name.")

    def setPenColor(self, color):
        """setPenColor(color)
        Set the color of the pen. Color can be anything that can be passed to
        Qcolor().
        """
        pen = QtGui.QPen()
        if isinstance(color, tuple):
            pen.setColor(QtGui.QColor(*color))
        else:
            pen.setColor(QtGui.QColor(color))
        self._painter.setPen(pen)

    def addLayer(self, overlay, x=0, y=0):
        """addOverlay(overlay, x=0, y=0)
        Add an overlay icon to the icon (add the specified position).
        """
        pm = self._getPixmap(overlay)
        self._painter.drawPixmap(x, y, pm)

    def addLine(self, x1, y1, x2, y2):
        """addLine( x1, y1, x2, y2)
        Add a line to the icon.
        """
        self._painter.drawLine(x1, y1, x2, y2)

    def addPoint(self, x, y):
        """addPoint( x, y)
        Add a point to the icon.
        """
        self._painter.drawPoint(x, y)

    def addMenuArrow(self, strength=100):
        """addMenuArrow()
        Adds a menu arrow to the icon to let the user know the icon
        is clickable.
        """
        x, y = 0, 12
        a1, a2 = int(strength / 2), strength
        # Zeroth line of 3+2
        self.setPenColor((0, 0, 0, a1))
        self.addPoint(x + 0, y - 1)
        self.addPoint(x + 4, y - 1)
        self.setPenColor((0, 0, 0, a2))
        self.addPoint(x + 1, y - 1)
        self.addPoint(x + 2, y - 1)
        self.addPoint(x + 3, y - 1)
        # First line of 3+2
        self.setPenColor((0, 0, 0, a1))
        self.addPoint(x + 0, y + 0)
        self.addPoint(x + 4, y + 0)
        self.setPenColor((0, 0, 0, a2))
        self.addPoint(x + 1, y + 0)
        self.addPoint(x + 2, y + 0)
        self.addPoint(x + 3, y + 0)
        # Second line of 3
        self.addPoint(x + 1, y + 1)
        self.addPoint(x + 2, y + 1)
        self.addPoint(x + 3, y + 1)
        # Third line of 1+2
        self.addPoint(x + 2, y + 2)
        self.setPenColor((0, 0, 0, a1))
        self.addPoint(x + 1, y + 2)
        self.addPoint(x + 3, y + 2)
        # Fourth line of 1
        self.setPenColor((0, 0, 0, a2))
        self.addPoint(x + 2, y + 3)


# todo: not used; remove me?
class TabCloseButton(QtWidgets.QToolButton):
    """TabCloseButton

    This class implements a very compact close button to be used in tabs.
    It allows managing tab (the closing part of it) in a fast and intuitive
    fashion.

    """

    SIZE = 5, 8

    def __init__(self):
        QtWidgets.QToolButton.__init__(self)

        # Init
        self.setIconSize(QtCore.QSize(*self.SIZE))
        self.setStyleSheet("QToolButton{ border:none; padding:0px; margin:0px; }")
        self.setIcon(self.getCrossIcon1())

    def mousePressEvent(self, event):
        # Get tabs
        tabs = self.parent().parent()
        # Get index from position
        pos = self.mapTo(tabs, event.pos())
        index = tabs.tabBar().tabAt(pos)
        # Close it
        tabs.tabCloseRequested.emit(index)

    def enterEvent(self, event):
        QtWidgets.QToolButton.enterEvent(self, event)
        self.setIcon(self.getCrossIcon2())

    def leaveEvent(self, event):
        QtWidgets.QToolButton.leaveEvent(self, event)
        self.setIcon(self.getCrossIcon1())

    def _createCrossPixmap(self, alpha):
        artist = IconArtist(self.SIZE)
        #
        artist.setPenColor((0, 0, 0, alpha))
        #
        artist.addPoint(0, 0)
        artist.addPoint(1, 1)
        artist.addPoint(2, 2)
        artist.addPoint(3, 3)
        artist.addPoint(4, 4)
        artist.addPoint(0, 4)
        artist.addPoint(1, 3)
        artist.addPoint(3, 1)
        artist.addPoint(4, 0)
        #
        artist.setPenColor((0, 0, 0, int(0.5 * alpha)))
        #
        artist.addPoint(1, 0)
        artist.addPoint(0, 1)
        artist.addPoint(2, 1)
        artist.addPoint(1, 2)
        artist.addPoint(3, 2)
        artist.addPoint(2, 3)
        artist.addPoint(4, 3)
        artist.addPoint(3, 4)
        #
        artist.addPoint(0, 3)
        artist.addPoint(1, 4)
        artist.addPoint(3, 0)
        artist.addPoint(4, 1)
        #
        return artist.finish().pixmap(*self.SIZE)

    def getCrossIcon1(self):
        if hasattr(self, "_cross1"):
            pm = self._cross1
        else:
            pm = self._createCrossPixmap(80)
        return QtGui.QIcon(pm)

    def getCrossIcon2(self):
        if hasattr(self, "_cross2"):
            pm = self._cross2
        else:
            pm = self._createCrossPixmap(240)
        # Set
        return QtGui.QIcon(pm)


# todo: not used; remove me?
class ToolButtonWithMenuIndication(QtWidgets.QToolButton):
    """ToolButtonWithMenuIndication

    Tool button that wraps the icon in a slightly larger icon that
    contains a small arrow that lights up when hovering over the icon.

    The button itself is not drawn. If the icon is clicked, the
    customContextMenuRequested signal of the "grandparent" is emitted. In
    this way we realize a suble icon that can be clicked on to show a menu.

    """

    SIZE = 21, 16

    def __init__(self):
        QtWidgets.QToolButton.__init__(self)

        # Init
        self.setIconSize(QtCore.QSize(*self.SIZE))
        self.setStyleSheet("QToolButton{ border: none; }")

        # Create arrow pixmaps
        self._menuarrow1 = self._createMenuArrowPixmap(0)
        self._menuarrow2 = self._createMenuArrowPixmap(70)
        self._menuarrow = self._menuarrow1

        # Variable to keep icon
        self._icon = None

        # Variable to keep track of when the mouse was pressed, so that
        # we can allow dragging as well as clicking the menu.
        self._menuPressed = False

    def mousePressEvent(self, event):
        # Ignore event so that the tabbar will change to that tab
        event.ignore()
        self._menuPressed = event.pos()

    def mouseMoveEvent(self, event):
        QtWidgets.QToolButton.mouseMoveEvent(self, event)
        if self._menuPressed:
            dragDist = QtWidgets.QApplication.startDragDistance()
            if (event.pos() - self._menuPressed).manhattanLength() >= dragDist:
                self._menuPressed = False

    def mouseReleaseEvent(self, event):
        event.ignore()
        if self._menuPressed:
            tabs = self.parent().parent()
            pos = self.mapTo(tabs, event.pos())
            tabs.customContextMenuRequested.emit(pos)

    def enterEvent(self, event):
        QtWidgets.QToolButton.enterEvent(self, event)
        self._menuarrow = self._menuarrow2
        self.setIcon()
        self._menuPressed = False

    def leaveEvent(self, event):
        QtWidgets.QToolButton.leaveEvent(self, event)
        self._menuarrow = self._menuarrow1
        self.setIcon()
        self._menuPressed = False

    def setIcon(self, icon=None):

        # Store icon if given, otherwise use buffered version
        if icon is not None:
            self._icon = icon

        # Compose icon by superimposing the menuarrow pixmap
        artist = IconArtist(self.SIZE)
        if self._icon:
            artist.addLayer(self._icon, 5, 0)
        artist.addLayer(self._menuarrow, 0, 0)
        icon = artist.finish()

        # Set icon
        QtWidgets.QToolButton.setIcon(self, icon)

    def _createMenuArrowPixmap(self, strength):
        artist = IconArtist()
        artist.addMenuArrow(strength)
        return artist.finish().pixmap(16, 16)


class TabToolButton(QtWidgets.QToolButton):
    """TabToolButton

    Base menu for editor and shell tabs.

    """

    SIZE = 16, 16

    def __init__(self, *args):
        QtWidgets.QToolButton.__init__(self, *args)

        # Init
        self.setIconSize(QtCore.QSize(*self.SIZE))
        self.setStyleSheet("QToolButton{ border: none; }")

    def mousePressEvent(self, event):
        # Ignore event so that the tabbar will change to that tab
        event.ignore()


class TabToolButtonWithCloseButton(TabToolButton):
    """TabToolButtonWithCloseButton

    Tool button that wraps the icon in a slightly larger icon that
    contains a small cross that can be used to invoke a close request.

    """

    SIZE = 22, 16
    CROSS_OFFSET = 0, 2

    def __init__(self, *args):
        TabToolButton.__init__(self, *args)

        # Variable to keep icon
        self._icon = None
        self._cross = self.getCrossPixmap1()

        # For mouse tracking inside icon
        self.setMouseTracking(True)
        self._overCross = False

    def _isOverCross(self, pos):
        x1, x2 = self.CROSS_OFFSET[0], self.CROSS_OFFSET[0] + 5 + 1
        y1, y2 = self.CROSS_OFFSET[1], self.CROSS_OFFSET[1] + 5 + 1
        if pos.x() >= x1 and pos.x() <= x2 and pos.y() >= y1 and pos.y() <= y2:
            return True
        else:
            return False

    def mousePressEvent(self, event):
        if self._isOverCross(event.pos()):
            # Accept event so that the tabbar will NOT change to that tab
            event.accept()
        else:
            event.ignore()

    def mouseReleaseEvent(self, event):
        if self._isOverCross(event.pos()):
            event.accept()
            # Get tabs
            tabs = self.parent().parent()
            # Get index from position
            pos = self.mapTo(tabs, event.pos())
            index = tabs.tabBar().tabAt(pos)
            # Close it
            tabs.tabCloseRequested.emit(index)
        else:
            event.ignore()

    def mouseMoveEvent(self, event):
        QtWidgets.QToolButton.mouseMoveEvent(self, event)
        new_overCross = self._isOverCross(event.pos())
        if new_overCross != self._overCross:
            self._overCross = new_overCross
            if new_overCross:
                self._cross = self.getCrossPixmap2()
            else:
                self._cross = self.getCrossPixmap1()
            self.setIcon()

    def leaveEvent(self, event):
        if self._overCross:
            self._overCross = False
            self._cross = self.getCrossPixmap1()
            self.setIcon()

    def setIcon(self, icon=None):

        # Store icon if given, otherwise use buffered version
        if icon is not None:
            self._icon = icon

        # Compose icon by superimposing the menuarrow pixmap
        artist = IconArtist(self.SIZE)
        if self._icon:
            if self.CROSS_OFFSET[0] > 8:
                artist.addLayer(self._icon, 0, 0)
            else:
                artist.addLayer(self._icon, 6, 0)
        artist.addLayer(self._cross, *self.CROSS_OFFSET)
        icon = artist.finish()

        # Set icon
        QtWidgets.QToolButton.setIcon(self, icon)

    def _createMenuArrowPixmap(self, strength):
        artist = IconArtist()
        artist.addMenuArrow(strength)
        return artist.finish().pixmap(16, 16)

    def _createCrossPixmap(self, alpha):
        artist = IconArtist((5, 5))
        #
        artist.setPenColor((0, 0, 0, alpha))
        #
        artist.addPoint(0, 0)
        artist.addPoint(1, 1)
        artist.addPoint(2, 2)
        artist.addPoint(3, 3)
        artist.addPoint(4, 4)
        artist.addPoint(0, 4)
        artist.addPoint(1, 3)
        artist.addPoint(3, 1)
        artist.addPoint(4, 0)
        #
        artist.setPenColor((0, 0, 0, int(0.5 * alpha)))
        #
        artist.addPoint(1, 0)
        artist.addPoint(0, 1)
        artist.addPoint(2, 1)
        artist.addPoint(1, 2)
        artist.addPoint(3, 2)
        artist.addPoint(2, 3)
        artist.addPoint(4, 3)
        artist.addPoint(3, 4)
        #
        artist.addPoint(0, 3)
        artist.addPoint(1, 4)
        artist.addPoint(3, 0)
        artist.addPoint(4, 1)
        #
        return artist.finish().pixmap(5, 5)

    def getCrossPixmap1(self):
        if hasattr(self, "_cross1"):
            pm = self._cross1
        else:
            pm = self._createCrossPixmap(50)
        return pm

    def getCrossPixmap2(self):
        if hasattr(self, "_cross2"):
            pm = self._cross2
        else:
            pm = self._createCrossPixmap(240)
        # Set
        return pm


class EditorTabToolButton(TabToolButtonWithCloseButton):
    """Button for the tabs of the editors. This is just a
    tight wrapper for the icon.
    """

    def updateIcon(self, isDirty, isMain, isPinned, nBlocks=10001):

        # Init drawing
        artist = IconArtist()

        # Create base
        if isDirty:
            artist.addLayer("page_white_dirty")
            artist.setPenColor("#f00")
        else:
            artist.addLayer("page_white")
            artist.setPenColor("#444")

        # Paint lines
        if not nBlocks:
            nLines = 0
        elif nBlocks <= 10:
            nLines = 1
        elif nBlocks <= 100:
            nLines = 2
        elif nBlocks <= 1000:
            nLines = 3
        elif nBlocks <= 10000:
            nLines = 4
        else:
            nLines = 5
        #
        fraction = float(nBlocks) / 10**nLines
        fraction = min(fraction, 1.0)
        #
        for i in range(nLines):
            y = 4 + 2 * i
            n = 5
            if y > 6:
                n = 8
            # if i == nLines-1:
            #    n = int(fraction * n)
            artist.addLine(4, y, 4 + n, y)

        # Overlays
        if isMain:
            artist.addLayer("overlay_star")
        if isPinned:
            artist.addLayer("overlay_thumbnail")
        if isDirty:
            artist.addLayer("overlay_disk")

        # Apply
        self.setIcon(artist.finish())


class ShellIconMaker:
    """Object that can make an icon for the shells"""

    POSITION = (6, 7)  # absolute position of center of wheel.

    # Relative position for the wheel at two levels. Center is at (3,,3)
    POSITIONS1 = [(2, 2), (3, 2), (4, 2), (4, 3), (4, 4), (3, 4), (2, 4), (2, 3)]
    POSITIONS2 = [
        (2, 1),
        (3, 1),
        (4, 1),
        (5, 2),
        (5, 3),
        (5, 4),
        (4, 5),
        (3, 5),
        (2, 5),
        (1, 4),
        (1, 3),
        (1, 2),
    ]

    # Maps to make transitions between levels more natural
    MAP1to2 = [1, 2, 4, 5, 7, 8, 10, 11]
    MAP2to1 = [1, 2, 3, 3, 4, 5, 5, 6, 7, 7, 0, 1]

    MAX_ITERS_IN_LEVEL_1 = 2

    def __init__(self, objectWithIcon):

        self._objectWithIcon = objectWithIcon

        # Motion properties
        self._index = 0
        self._level = 0
        self._count = 0  #  to count number of iters in level 1

        # Prepare blob pixmap
        self._blob = self._createBlobPixmap()
        self._legs = self._createLegsPixmap()

        # Create timer
        self._timer = QtCore.QTimer(None)
        self._timer.setInterval(200)
        self._timer.setSingleShot(False)
        self._timer.timeout.connect(self.onTimer)

    def setIcon(self, icon):
        self._objectWithIcon.setIcon(icon)

    def _createBlobPixmap(self):

        artist = IconArtist()
        artist.setPenColor((0, 150, 0, 255))
        artist.addPoint(1, 1)
        artist.setPenColor((0, 150, 0, 200))
        artist.addPoint(1, 0)
        artist.addPoint(1, 2)
        artist.addPoint(0, 1)
        artist.addPoint(2, 1)
        artist.setPenColor((0, 150, 0, 100))
        artist.addPoint(0, 0)
        artist.addPoint(2, 0)
        artist.addPoint(0, 2)
        artist.addPoint(2, 2)
        return artist.finish().pixmap(16, 16)

    def _createLegsPixmap(self):
        artist = IconArtist()
        x, y = self.POSITION
        artist.setPenColor((0, 50, 0, 150))
        artist.addPoint(x + 1, y - 1)
        artist.addPoint(x + 1, y - 2)
        artist.addPoint(x + 0, y - 2)
        artist.addPoint(x + 3, y + 1)
        artist.addPoint(x + 4, y + 1)
        artist.addPoint(x + 4, y + 2)
        artist.addPoint(x + 2, y + 3)
        artist.addPoint(x + 2, y + 4)
        artist.addPoint(x + 0, y + 3)
        artist.addPoint(x + 0, y + 4)
        artist.addPoint(x - 1, y + 2)
        artist.addPoint(x - 2, y + 2)
        artist.addPoint(x - 1, y + 0)
        artist.addPoint(x - 2, y + 0)
        return artist.finish().pixmap(16, 16)

    def updateIcon(self, status="Ready"):
        """updateIcon(status)
        Public method to set what state the icon must show.
        """

        # Normalize and store
        if isinstance(status, str):
            status = status.lower()
        self._status = status

        # Handle
        if status == "busy":
            self._index = 0
            if self._level == 2:
                self._index = self.MAP2to1[self._index]
            self._level = 1

        elif status == "very busy":
            self._index = 0
            if self._level == 1:
                self._index = self.MAP1to2[self._index]
            self._level = 2

        else:
            self._level = 0

        # At least one timer iteration
        self._timer.start()

    def _nextIndex(self):
        self._index += 1
        if self._level == 1 and self._index >= 8:
            self._index = 0
        elif self._level == 2 and self._index >= 12:
            self._index = 0

    def _index1(self):
        return self._index

    def _index2(self):
        n = [0, 8, 12][self._level]
        index = self._index + n / 2
        if index >= n:
            index -= n
        return int(index)

    def onTimer(self):
        """onTimer()
        Invoked on each timer iteration. Will call the static drawing
        methods if in level 0. Otherwise will invoke drawInMotion().
        This method also checks if we should change levels and calculates
        how this is best achieved.
        """
        if self._level == 0:
            # Turn of timer
            self._timer.stop()
            # Draw
            if self._status in ["ready", "more"]:
                self.drawReady()
            elif self._status == "debug":
                self.drawDebug()
            elif self._status == "dead":
                self.drawDead()
            else:
                self.drawDead()

        elif self._level == 1:
            # Draw
            self.drawInMotion()
            # Next, this is always intermediate
            self._nextIndex()
            self._count += 1

        elif self._level == 2:
            # Draw
            self.drawInMotion()
            # Next
            self._nextIndex()

    def drawReady(self):
        """drawReady()
        Draw static icon for when in ready mode.
        """
        artist = IconArtist("application")
        artist.addLayer(self._blob, *self.POSITION)
        self.setIcon(artist.finish())

    def drawDebug(self):
        """drawDebug()
        Draw static icon for when in debug mode.
        """
        artist = IconArtist("application")
        artist.addLayer(self._blob, *self.POSITION)
        artist.addLayer(self._legs)
        self.setIcon(artist.finish())

    def drawDead(self):
        """drawDead()
        Draw static empty icon for when the kernel is dead.
        """
        artist = IconArtist("application")
        self.setIcon(artist.finish())

    def drawInMotion(self):
        """drawInMotion()
        Draw one frame of the icon in motion. Position of the blobs
        is determined from the index and the list of locations.
        """

        # Init drawing
        artist = IconArtist("application")

        # Define params
        dx, dy = self.POSITION[0] - 3, self.POSITION[1] - 3
        blob = self._blob
        #
        if self._level == 1:
            positions = self.POSITIONS1
        elif self._level == 2:
            positions = self.POSITIONS2

        # Draw
        pos1 = positions[self._index1()]
        pos2 = positions[self._index2()]
        artist.addLayer(blob, pos1[0] + dx, pos1[1] + dy)
        artist.addLayer(blob, pos2[0] + dx, pos2[1] + dy)

        # Done
        self.setIcon(artist.finish())



#
# Copyright © 2014-2015 Colin Duquesnoy
# Copyright © 2009- The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""
Provides QtCore classes and functions.
"""
from . import PYQT6, PYQT5, PYSIDE2, PYSIDE6, PythonQtError


if PYQT6:
    from PyQt6 import QtCore
    from PyQt6.QtCore import *
    from PyQt6.QtCore import pyqtSignal as Signal
    from PyQt6.QtCore import pyqtBoundSignal as SignalInstance
    from PyQt6.QtCore import pyqtSlot as Slot
    from PyQt6.QtCore import pyqtProperty as Property
    from PyQt6.QtCore import QT_VERSION_STR as __version__

    # For issue #153
    from PyQt6.QtCore import QDateTime

    QDateTime.toPython = QDateTime.toPyDateTime

    # Map missing methods
    QCoreApplication.exec_ = QCoreApplication.exec
    QEventLoop.exec_ = QEventLoop.exec
    QThread.exec_ = QThread.exec

    QLibraryInfo.location = QLibraryInfo.path

    # Those are imported from `import *`
    del pyqtSignal, pyqtBoundSignal, pyqtSlot, pyqtProperty, QT_VERSION_STR

    # Allow unscoped access for enums inside the QtCore module
    from .enums_compat import promote_enums

    promote_enums(QtCore)
    del QtCore
elif PYQT5:
    from PyQt5.QtCore import *
    from PyQt5.QtCore import pyqtSignal as Signal
    from PyQt5.QtCore import pyqtBoundSignal as SignalInstance
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtCore import pyqtProperty as Property
    from PyQt5.QtCore import QT_VERSION_STR as __version__

    # For issue #153
    from PyQt5.QtCore import QDateTime

    QDateTime.toPython = QDateTime.toPyDateTime

    # Those are imported from `import *`
    del pyqtSignal, pyqtBoundSignal, pyqtSlot, pyqtProperty, QT_VERSION_STR

elif PYSIDE6:
    from PySide6.QtCore import *
    import PySide6.QtCore

    __version__ = PySide6.QtCore.__version__

    # obsolete in qt6
    Qt.BackgroundColorRole = Qt.BackgroundRole
    Qt.TextColorRole = Qt.ForegroundRole
    Qt.MidButton = Qt.MiddleButton

    # Map DeprecationWarning methods
    QCoreApplication.exec_ = QCoreApplication.exec
    QEventLoop.exec_ = QEventLoop.exec
    QThread.exec_ = QThread.exec
    QTextStreamManipulator.exec_ = QTextStreamManipulator.exec

elif PYSIDE2:
    from PySide2.QtCore import *

    try:  # may be limited to PySide-5.11a1 only
        from PySide2.QtGui import QStringListModel
    except Exception:
        pass

    import PySide2.QtCore

    __version__ = PySide2.QtCore.__version__
else:
    raise PythonQtError("No Qt bindings could be found")



""" Setup script for the Pyzo package.

Notes on how to do a release. Mostly for my own convenience:

* Write release notes
* Bump `__version__`
* `git tag vx.y.z`
* `git push vx.y.z` (makes Azure build the binaries and push to a GH release)
* Update links on Pyzo website
* `python setup.py sdist upload`

"""

import os
import sys

try:
    from setuptools import find_packages, setup
except ImportError:
    from distutils.core import setup


def get_version_and_doc(filename):
    NS = dict(__version__="", __doc__="")
    docStatus = 0  # Not started, in progress, done
    for line in open(filename, "rb").read().decode().splitlines():
        if line.startswith("__version__"):
            exec(line.strip(), NS, NS)
        elif line.startswith('"""'):
            if docStatus == 0:
                docStatus = 1
                line = line.lstrip('"')
            elif docStatus == 1:
                docStatus = 2
        if docStatus == 1:
            NS["__doc__"] += line.rstrip() + "\n"
    if not NS["__version__"]:
        raise RuntimeError("Could not find __version__")
    return NS["__version__"], NS["__doc__"]


version, doc = get_version_and_doc(
    os.path.join(os.path.dirname(__file__), "pyzo", "__init__.py")
)


setup(
    name="pyzo",
    version=version,
    description="the Python IDE for scientific computing",
    long_description=doc,
    author="Almar Klein",
    author_email="almar.klein@gmail.com",
    license="2-Clause BSD",
    url="https://pyzo.org",
    keywords="Python interactive IDE Qt science computing",
    platforms="any",
    provides=["pyzo"],
    python_requires=">=3.5.0",
    install_requires=[
        "packaging"
    ],  # and 'PySide2' or 'PyQt5' (less sure about PySide/PyQt4)
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_dir={"pyzo": "pyzo"},
    package_data={
        "pyzo": [
            "resources/*.*",
            "resources/icons/*.*",
            "resources/appicons/*.*",
            "resources/images/*.*",
            "resources/fonts/*.*",
            "resources/themes/*.*",
            "resources/translations/*.*",
        ]
    },
    data_files=[
        ("", ["README.md", "LICENSE.md", "pyzo.appdata.xml", "pyzolauncher.py"])
    ],
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={
        "console_scripts": [
            "pyzo = pyzo.__main__:main",
        ],
    },
)


# Post processing:
# Install appdata.xml on Linux if we are installing in the system Python
if sys.platform.startswith("linux") and sys.prefix.startswith("/usr"):
    if len(sys.argv) >= 2 and sys.argv[1] == "install":
        fname = "pyzo.appdata.xml"
        filename1 = os.path.join(os.path.dirname(__file__), fname)
        filename2 = os.path.join("/usr/share/metainfo", fname)
        try:
            bb = open(filename1, "rb").read()
            open(filename2, "wb").write(bb)
        except PermissionError:
            pass  # No sudo, no need to warn
        except Exception as err:
            print("Could not install %s: %s" % (fname, str(err)))
        else:
            print("Installed %s" % fname)



# -*- coding: utf-8 -*-
# Copyright (C) 2018, the codeeditor development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

from . import Parser, BlockState, text_type

# Import tokens in module namespace
from .tokens import (
    CommentToken,
    StringToken,
    UnterminatedStringToken,
    IdentifierToken,
    ClassNameToken,
    KeywordToken,
    NumberToken,
    OpenParenToken,
    CloseParenToken,
)


class SExprParser(Parser):
    """Parser for S-expressions."""

    _extensions = [".lisp", ".ss", ".sls", ".scm"]

    _keywords = []  # can be overloaded

    def parseLine(self, line, comment_level=0):
        """parseLine(line, comment_level=0)

        Parse a line of code, yielding tokens.
        previousstate is the state of the previous block, and is used
        to handle line continuation and multiline strings.

        """
        line = text_type(line)

        if comment_level < 0:
            comment_level = 0
        if comment_level > 0:
            token = CommentToken(line, 0, 0)

        pos = 0
        while pos < len(line):
            pos = self._skip_whitespace(line, pos)
            if pos >= len(line):
                break

            # Parse block comments
            if line[pos] == "(" and pos < len(line) - 1 and line[pos + 1] == ";":
                if comment_level == 0:
                    token = CommentToken(line, pos, pos)
                comment_level += 1
                pos += 1
            elif line[pos] == ";" and pos < len(line) - 1 and line[pos + 1] == ")":
                if comment_level == 1:
                    token.end = pos + 2
                    yield token
                comment_level = max(0, comment_level - 1)
                pos += 1

            elif comment_level > 0:
                pos += 1

            else:
                # Outside of block comments ...

                if line[pos] == ";" and pos < len(line) - 1 and line[pos + 1] == ";":
                    yield CommentToken(line, pos, len(line))
                    pos = len(line)
                elif line[pos] == "(":
                    token = OpenParenToken(line, pos, pos + 1)
                    token._style = "("
                    yield token
                    pos += 1
                elif line[pos] == ")":
                    token = CloseParenToken(line, pos, pos + 1)
                    token._style = ")"
                    yield token
                    pos += 1
                elif line[pos] == '"':
                    i0 = pos
                    esc = False
                    for i in range(i0 + 1, len(line)):
                        if not esc and line[i] == '"':
                            pos = i + 1
                            yield StringToken(line, i0, pos)
                            break
                        esc = line[i] == "\\"
                    else:
                        yield UnterminatedStringToken(line, i0, len(line))
                        pos = len(line)
                else:
                    # word: number, keyword or normal identifier
                    i0 = pos
                    for i in range(i0, len(line)):
                        if line[i] in " \t\r\n)":
                            yield self._get_token_for_word(line, i0, i)
                            pos = i
                            break
                    else:
                        pos = len(line)
                        yield self._get_token_for_word(line, i0, len(line))

        if comment_level > 0:
            token.end = len(line)
            yield token
        yield BlockState(comment_level)

    def _skip_whitespace(self, line, pos):
        while pos < len(line):
            if line[pos] not in " \t\r\n":
                break
            pos += 1
        return pos

    def _get_token_for_word(self, line, i0, i1):
        word = line[i0:i1]
        is_number = False
        try:
            float(word)
            is_number = True
        except ValueError:
            pass

        if is_number or word.startswith("$"):
            return NumberToken(line, i0, i1)
        elif word in self._keywords:  # highlight extra
            return ClassNameToken(line, i0, i1)  # ClassNameToken or FunctionNameToken
        elif i0 > 0 and line[i0 - 1] == "(":  # First element in expression is "keyword"
            return KeywordToken(line, i0, i1)
        else:
            return IdentifierToken(line, i0, i1)


class WatParser(SExprParser):
    """Parser for textual WASM (WAT) code."""

    _extensions = [".wat", ".wast"]

    _keywords = [
        "module",
        "type",
        "import",
        "func",
        "table",
        "memory",
        "global",
        "export",
        "start",
        "element",
        "data",
    ]



# -*- coding: utf-8 -*-
# Copyright (C) 2016, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" Module editor

Defines the PyzoEditor class which is used to edit documents.
This module/class also implements all the relatively low level
file loading/saving /reloading stuff.

"""

import os, sys
import re, codecs

from pyzo.qt import QtCore, QtGui, QtWidgets

qt = QtGui

from pyzo.codeeditor import Manager
from pyzo.core.menu import EditorContextMenu
from pyzo.core.baseTextCtrl import BaseTextCtrl, normalizePath
from pyzo.core.pyzoLogging import print  # noqa
import pyzo


# Set default line ending (if not set)
if not pyzo.config.settings.defaultLineEndings:
    if sys.platform.startswith("win"):
        pyzo.config.settings.defaultLineEndings = "CRLF"
    else:
        pyzo.config.settings.defaultLineEndings = "LF"


def determineEncoding(bb):
    """Get the encoding used to encode a file.
    Accepts the bytes of the file. Returns the codec name. If the
    codec could not be determined, uses UTF-8.
    """

    # Init
    firstTwoLines = bb.split(b"\n", 2)[:2]
    encoding = "UTF-8"

    for line in firstTwoLines:

        # Try to make line a string
        try:
            line = line.decode("ASCII").strip()
        except Exception:
            continue

        # Has comment?
        if line and line[0] == "#":

            # Matches regular expression given in PEP 0263?
            expression = r"coding[:=]\s*([-\w.]+)"
            result = re.search(expression, line)
            if result:

                # Is it a known encoding? Correct name if it is
                candidate_encoding = result.group(1)
                try:
                    c = codecs.lookup(candidate_encoding)
                    candidate_encoding = c.name
                except Exception:
                    pass
                else:
                    encoding = candidate_encoding

    # Done
    return encoding


def determineLineEnding(text):
    """Get the line ending style used in the text.
    \n, \r, \r\n,
    The EOLmode is determined by counting the occurrences of each
    line ending...
    """
    text = text[:32768]  # Limit search for large files

    # test line ending by counting the occurrence of each
    c_win = text.count("\r\n")
    c_mac = text.count("\r") - c_win
    c_lin = text.count("\n") - c_win
    # set the appropriate style
    if c_win > c_mac and c_win > c_lin:
        mode = "\r\n"
    elif c_mac > c_win and c_mac > c_lin:
        mode = "\r"
    else:
        mode = "\n"

    # return
    return mode


def determineIndentation(text):
    return determineIndentationAndTrailingWS(text)[0]


def determineIndentationAndTrailingWS(text):
    """Get the indentation used in this document and whether there is
    any trailing whitespace.
    The text is analyzed to find the most used indentations.
    The result is -1 if tab indents are most common.
    A positive result means spaces are used; the amount
    signifies the amount of spaces per indentation.
    0 is returned if the indentation could not be determined.
    The second return value is the number of lines with trailing ws.
    """

    text = text[:32768]  # Limit search for large files

    # create dictionary of indents, -1 means a tab
    indents = {}
    indents[-1] = 0
    trailing = 0

    lines = text.splitlines()
    lines.insert(0, "")  # so the lines start at 1
    for i in range(len(lines)):
        line = lines[i]

        # remove indentation
        lineA = line.lstrip()
        lineB = line.rstrip()
        lineC = lineA.rstrip()
        indent = len(line) - len(lineA)
        if len(lineB) < len(line):
            trailing += 1

        line = lineC
        if line.startswith("#"):
            continue
        else:
            # remove everything after the #
            line = line.split("#", 1)[0].rstrip()
        if not line:
            # continue of no line left
            continue

        # a colon means there will be an indent
        # check the next line (or the one thereafter)
        # and calculate the indentation difference with THIS line.
        if line.endswith(":"):
            if len(lines) > i + 2:
                line2 = lines[i + 1]
                tmp = line2.lstrip()
                if not tmp:
                    line2 = lines[i + 2]
                    tmp = line2.lstrip()
                if tmp:
                    ind2 = len(line2) - len(tmp)
                    ind3 = ind2 - indent
                    if line2.startswith("\t"):
                        indents[-1] += 1
                    elif ind3 > 0:
                        if ind3 not in indents:
                            indents[ind3] = 1
                        indents[ind3] += 1

    # find which was the most common tab width.
    indent, maxvotes = 0, 0
    for nspaces in indents:
        if indents[nspaces] > maxvotes:
            indent, maxvotes = nspaces, indents[nspaces]
    # print "found tabwidth %i" % indent
    return indent, trailing


# To give each new file a unique name
newFileCounter = 0


def createEditor(parent, filename=None):
    """Tries to load the file given by the filename and
    if succesful, creates an editor instance to put it in,
    which is returned.
    If filename is None, an new/unsaved/temp file is created.
    """

    if filename is None:

        # Increase counter
        global newFileCounter
        newFileCounter += 1

        # Create editor
        editor = PyzoEditor(parent)
        editor.document().setModified(True)
        editor.removeTrailingWS = True

        # Set name
        editor._name = "<tmp {}>".format(newFileCounter)

    else:

        # check and normalize
        if not os.path.isfile(filename):
            raise IOError("File does not exist '%s'." % filename)

        # load file (as bytes)
        with open(filename, "rb") as f:
            bb = f.read()
            f.close()

        # convert to text, be gentle with files not encoded with utf-8
        encoding = determineEncoding(bb)
        text = bb.decode(encoding, "replace")

        # process line endings
        lineEndings = determineLineEnding(text)

        # if we got here safely ...

        # create editor and set text
        editor = PyzoEditor(parent)
        editor.setPlainText(text)
        editor.lineEndings = lineEndings
        editor.encoding = encoding
        editor.document().setModified(False)

        # store name and filename
        editor._filename = filename
        editor._name = os.path.split(filename)[1]

        # process indentation and trailing
        indentWidth, trailing = determineIndentationAndTrailingWS(text)
        editor.removeTrailingWS = trailing < 10  # not too much for ugly diffs
        if indentWidth == -1:  # Tabs
            editor.setIndentWidth(pyzo.config.settings.defaultIndentWidth)
            editor.setIndentUsingSpaces(False)
        elif indentWidth:
            editor.setIndentWidth(indentWidth)
            editor.setIndentUsingSpaces(True)

    if editor._filename:
        editor._modifyTime = os.path.getmtime(editor._filename)

    # Set parser
    if editor._filename:
        ext = os.path.splitext(editor._filename)[1]
        parser = Manager.suggestParser(ext, text)
        editor.setParser(parser)
    else:
        # todo: rename style -> parser
        editor.setParser(pyzo.config.settings.defaultStyle)

    # return
    return editor


class PyzoEditor(BaseTextCtrl):

    # called when dirty changed or filename changed, etc
    somethingChanged = QtCore.Signal()

    def __init__(self, parent, **kwds):
        super().__init__(parent, showLineNumbers=True, **kwds)

        # Init filename and name
        self._filename = ""
        self._name = "<TMP>"

        # View settings
        # TODO: self.setViewWrapSymbols(view.showWrapSymbols)
        self.setShowLineEndings(pyzo.config.view.showLineEndings)
        self.setShowIndentationGuides(pyzo.config.view.showIndentationGuides)
        #
        self.setWrap(bool(pyzo.config.view.wrap))
        self.setHighlightCurrentLine(pyzo.config.view.highlightCurrentLine)
        self.setLongLineIndicatorPosition(pyzo.config.view.edgeColumn)
        # TODO: self.setFolding( int(view.codeFolding)*5 )
        # bracematch is set in baseTextCtrl, since it also applies to shells
        # dito for zoom and tabWidth

        # Set line endings to default
        self.lineEndings = pyzo.config.settings.defaultLineEndings

        # Set encoding to default
        self.encoding = "UTF-8"

        # Modification time to test file change
        self._modifyTime = 0

        self.modificationChanged.connect(self._onModificationChanged)

        # To see whether the doc has changed to update the parser.
        self.textChanged.connect(self._onModified)

        # This timer is used to hide the marker that shows which code is executed
        self._showRunCursorTimer = QtCore.QTimer()

        # Add context menu (the offset is to prevent accidental auto-clicking)
        self._menu = EditorContextMenu(self)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(
            lambda p: self._menu.popup(self.mapToGlobal(p) + QtCore.QPoint(0, 3))
        )

        # Update status bar
        self.cursorPositionChanged.connect(self._updateStatusBar)

    ## Properties

    @property
    def name(self):
        return self._name

    @property
    def filename(self):
        return self._filename

    @property
    def lineEndings(self):
        """
        Line-endings style of this file. Setter accepts machine-readable (e.g. '\r') and human-readable (e.g. 'CR') input
        """
        return self._lineEndings

    @lineEndings.setter
    def lineEndings(self, value):
        if value in ("\r", "\n", "\r\n"):
            self._lineEndings = value
            return
        try:
            self._lineEndings = {"CR": "\r", "LF": "\n", "CRLF": "\r\n"}[value]
        except KeyError:
            raise ValueError("Invalid line endings style %r" % value)

    @property
    def lineEndingsHumanReadable(self):
        """
        Current line-endings style, human readable (e.g. 'CR')
        """
        return {"\r": "CR", "\n": "LF", "\r\n": "CRLF"}[self.lineEndings]

    @property
    def encoding(self):
        """Encoding used to convert the text of this file to bytes."""
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        # Test given value, correct name if it exists
        try:
            c = codecs.lookup(value)
            value = c.name
        except Exception:
            value = codecs.lookup("UTF-8").name
        # Store
        self._encoding = value

    ##

    def justifyText(self):
        """Overloaded version of justifyText to make it use our
        configurable justificationwidth.
        """
        super().justifyText(pyzo.config.settings.justificationWidth)

    def showRunCursor(self, cursor):
        """
        Momentarily highlight a piece of code to show that this is being executed
        """

        extraSelection = QtWidgets.QTextEdit.ExtraSelection()
        extraSelection.cursor = cursor
        extraSelection.format.setBackground(QtCore.Qt.gray)
        self.setExtraSelections([extraSelection])

        self._showRunCursorTimer.singleShot(200, lambda: self.setExtraSelections([]))

    def id(self):
        """Get an id of this editor. This is the filename,
        or for tmp files, the name."""
        if self._filename:
            return self._filename
        else:
            return self._name

    def focusInEvent(self, event):
        """Test whether the file has been changed 'behind our back'"""
        # Act normally to the focus event
        BaseTextCtrl.focusInEvent(self, event)
        # Test file change
        self.testWhetherFileWasChanged()

    def testWhetherFileWasChanged(self):
        """testWhetherFileWasChanged()
        Test to see whether the file was changed outside our backs,
        and let the user decide what to do.
        Returns True if it was changed.
        """

        # get the path
        path = self._filename
        if not os.path.isfile(path):
            # file is deleted from the outside
            return

        # test the modification time...
        mtime = os.path.getmtime(path)
        if mtime != self._modifyTime:

            # ask user
            dlg = QtWidgets.QMessageBox(self)
            dlg.setWindowTitle("File was changed")
            dlg.setText(
                "File has been modified outside of the editor:\n" + self._filename
            )
            dlg.setInformativeText("Do you want to reload?")
            t = dlg.addButton("Reload", QtWidgets.QMessageBox.AcceptRole)  # 0
            dlg.addButton("Keep this version", QtWidgets.QMessageBox.RejectRole)  # 1
            dlg.setDefaultButton(t)

            # whatever the result, we will reset the modified time
            self._modifyTime = os.path.getmtime(path)

            # get result and act
            result = dlg.exec_()
            if result == QtWidgets.QMessageBox.AcceptRole:
                self.reload()
            else:
                pass  # when cancelled or explicitly said, do nothing

            # Return that indeed the file was changes
            return True

    def _onModificationChanged(self, changed):
        """Handler for the modificationChanged signal. Emit somethingChanged
        for the editorStack to update the modification notice."""
        self.somethingChanged.emit()

    def _onModified(self):
        pyzo.parser.parseThis(self)

    def _updateStatusBar(self):
        editor = pyzo.editors.getCurrentEditor()
        sb = pyzo.main.statusBar()
        sb.updateCursorInfo(editor)
        sb.updateFileEncodingInfo(editor)

    def dragMoveEvent(self, event):
        """Otherwise cursor can get stuck.
        https://bitbucket.org/iep-project/iep/issue/252
        https://qt-project.org/forums/viewthread/3180
        """
        if event.mimeData().hasUrls():
            event.acceptProposedAction()
        else:
            BaseTextCtrl.dropEvent(self, event)

    def dropEvent(self, event):
        """Drop files in the list."""
        if event.mimeData().hasUrls():
            # file: let the editorstack do the work.
            pyzo.editors.dropEvent(event)
        else:
            # text: act normal
            BaseTextCtrl.dropEvent(self, event)

    def showEvent(self, event=None):
        """Capture show event to change title."""
        # Act normally
        if event:
            BaseTextCtrl.showEvent(self, event)

        # Make parser update
        pyzo.parser.parseThis(self)

    def setTitleInMainWindow(self):
        """set the title  text in the main window to show filename."""

        # compose title
        name, path = self._name, self._filename
        if path:
            pyzo.main.setMainTitle(path)
        else:
            pyzo.main.setMainTitle(name)

    def save(self, filename=None):
        """Save the file. No checking is done."""

        # get filename
        if filename is None:
            filename = self._filename
        if not filename:
            raise ValueError("No filename specified, and no filename known.")

        # Test whether it was changed without us knowing. If so, dont save now.
        if self.testWhetherFileWasChanged():
            return

        # Remove whitespace in a single undo-able action
        if (
            self.removeTrailingWS
            or pyzo.config.settings.removeTrailingWhitespaceWhenSaving
        ):
            # Original cursor to put state back at the end
            oricursor = self.textCursor()
            # Screen cursor to select document
            screenCursor = self.textCursor()
            screenCursor.movePosition(screenCursor.Start)
            screenCursor.movePosition(screenCursor.End, screenCursor.KeepAnchor)
            # Cursor for doing the editor
            editCursor = self.textCursor()
            # Go!
            editCursor.beginEditBlock()
            try:
                editCursor.setPosition(screenCursor.selectionStart())
                editCursor.movePosition(editCursor.StartOfBlock)
                while (
                    editCursor.position() < screenCursor.selectionEnd()
                    or editCursor.position() <= screenCursor.selectionStart()
                ):
                    editCursor.movePosition(editCursor.StartOfBlock)
                    editCursor.movePosition(
                        editCursor.EndOfBlock, editCursor.KeepAnchor
                    )
                    text1 = editCursor.selectedText()
                    text2 = text1.rstrip()
                    if len(text1) != len(text2):
                        editCursor.insertText(text2)
                    if not editCursor.block().next().isValid():
                        break
                    editCursor.movePosition(editCursor.NextBlock)
            finally:
                self.setTextCursor(oricursor)
                editCursor.endEditBlock()

        # Get text and convert line endings
        text = self.toPlainText()
        text = text.replace("\n", self.lineEndings)

        # Make bytes
        bb = text.encode(self.encoding)

        # Store
        f = open(filename, "wb")
        try:
            f.write(bb)
        finally:
            f.close()

        # Update stats
        self._filename = normalizePath(filename)
        self._name = os.path.split(self._filename)[1]
        self.document().setModified(False)
        self._modifyTime = os.path.getmtime(self._filename)

        # update title (in case of a rename)
        self.setTitleInMainWindow()

        # allow item to update its texts (no need: onModifiedChanged does this)
        # self.somethingChanged.emit()

    def reload(self):
        """Reload text using the self._filename.
        We do not have a load method; we first try to load the file
        and only when we succeed create an editor to show it in...
        This method is only for reloading in case the file was changed
        outside of the editor."""

        # We can only load if the filename is known
        if not self._filename:
            return
        filename = self._filename

        # Remember where we are
        cursor = self.textCursor()
        linenr = cursor.blockNumber() + 1

        # Load file (as bytes)
        with open(filename, "rb") as f:
            bb = f.read()

        # Convert to text
        text = bb.decode("UTF-8")

        # Process line endings (before setting the text)
        self.lineEndings = determineLineEnding(text)

        # Set text
        self.setPlainText(text)
        self.document().setModified(False)

        # Go where we were (approximately)
        self.gotoLine(linenr)

    def deleteLines(self):
        cursor = self.textCursor()
        # Find start and end of selection
        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        # Expand selection: from start of first block to start of next block
        cursor.setPosition(start)
        cursor.movePosition(cursor.StartOfBlock)
        cursor.setPosition(end, cursor.KeepAnchor)
        cursor.movePosition(cursor.NextBlock, cursor.KeepAnchor)

        cursor.removeSelectedText()

    def duplicateLines(self):
        cursor = self.textCursor()
        # Find start and end of selection
        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        # Expand selection: from start of first block to start of next block
        cursor.setPosition(start)
        cursor.movePosition(cursor.StartOfBlock)
        cursor.setPosition(end, cursor.KeepAnchor)
        cursor.movePosition(cursor.NextBlock, cursor.KeepAnchor)

        text = cursor.selectedText()
        cursor.setPosition(start)
        cursor.movePosition(cursor.StartOfBlock)
        cursor.insertText(text)

    def commentCode(self):
        """
        Comment the lines that are currently selected
        """
        indents = []
        indentChar = " " if self.indentUsingSpaces() else "\t"

        def getIndent(cursor):
            text = cursor.block().text().rstrip()
            if text:
                indents.append(len(text) - len(text.lstrip()))

        def commentBlock(cursor):
            blockText = cursor.block().text()
            numMissingIndentChars = minindent - (
                len(blockText) - len(blockText.lstrip(indentChar))
            )
            if numMissingIndentChars > 0:
                # Prevent setPosition from leaving bounds of the current block
                # if there are too few indent characters (e.g. an empty line)
                cursor.insertText(indentChar * numMissingIndentChars)
            cursor.setPosition(cursor.block().position() + minindent)
            cursor.insertText("# ")

        self.doForSelectedBlocks(getIndent)
        minindent = min(indents) if indents else 0
        self.doForSelectedBlocks(commentBlock)

    def uncommentCode(self):
        """
        Uncomment the lines that are currently selected
        """
        # TODO: this should not be applied to lines that are part of a multi-line string

        # Define the uncomment function to be applied to all blocks
        def uncommentBlock(cursor):
            """
            Find the first # on the line; if there is just whitespace before it,
            remove the # and if it is followed by a space remove the space, too
            """
            text = cursor.block().text()
            commentStart = text.find("#")
            if commentStart == -1:
                return  # No comment on this line
            if text[:commentStart].strip() != "":
                return  # Text before the #
            # Move the cursor to the beginning of the comment
            cursor.setPosition(cursor.block().position() + commentStart)
            cursor.deleteChar()
            if text[commentStart:].startswith("# "):
                cursor.deleteChar()

        # Apply this function to all blocks
        self.doForSelectedBlocks(uncommentBlock)

    def toggleCommentCode(self):
        def toggleComment():
            """
            Toggles comments for the seclected text in editor, most of the code is
            taken from commentCode and uncommentCode
            """
            text_block = []

            def getBlocks(cursor):
                text = cursor.block().text()
                text_block.append(text)

            def commentBlock(cursor):
                cursor.setPosition(cursor.block().position())
                cursor.insertText("# ")

            def uncommentBlock(cursor):
                """
                Find the first # on the line; if there is just whitespace before it,
                remove the # and if it is followed by a space remove the space, too
                """
                cursor.setPosition(cursor.block().position())
                cursor.deleteChar()
                cursor.deleteChar()

            self.doForSelectedBlocks(getBlocks)
            commented = [item for item in text_block if item.startswith("# ")]
            if len(commented) == len(text_block):
                self.doForSelectedBlocks(uncommentBlock)
            else:
                self.doForSelectedBlocks(commentBlock)

        toggleComment()

    def gotoDef(self):
        """
        Goto the definition for the word under the cursor
        """

        # Get name of object to go to
        cursor = self.textCursor()
        if not cursor.hasSelection():
            cursor.select(cursor.WordUnderCursor)
        word = cursor.selection().toPlainText()

        # Send the open command to the shell
        s = pyzo.shells.getCurrentShell()
        if s is not None:
            if word and word.isidentifier():
                s.executeCommand("open %s\n" % word)
            else:
                s.write("Invalid identifier %r\n" % word)

    ## Introspection processing methods

    def processCallTip(self, cto):
        """Processes a calltip request using a CallTipObject instance."""
        # Try using buffer first
        if cto.tryUsingBuffer():
            return

        # Try obtaining calltip from the source
        sig = pyzo.parser.getFictiveSignature(cto.name, self, True)
        if sig:
            # Done
            cto.finish(sig)
        else:
            # Try the shell
            shell = pyzo.shells.getCurrentShell()
            if shell:
                shell.processCallTip(cto)

    def processAutoComp(self, aco):
        """Processes an autocomp request using an AutoCompObject instance."""

        # Try using buffer first
        if aco.tryUsingBuffer():
            return

        # Init name to poll by remote process (can be changed!)
        nameForShell = aco.name

        # Get normal fictive namespace
        fictiveNS = pyzo.parser.getFictiveNameSpace(self)
        fictiveNS = set(fictiveNS)

        # Add names
        if not aco.name:
            # "root" names
            aco.addNames(fictiveNS)
            # imports
            importNames, importLines = pyzo.parser.getFictiveImports(self)
            aco.addNames(importNames)
        else:
            # Prepare list of class names to check out
            classNames = [aco.name]
            handleSelf = True
            # Unroll supers
            while classNames:
                className = classNames.pop(0)
                if not className:
                    continue
                if handleSelf or (className in fictiveNS):
                    # Only the self list (only first iter)
                    fictiveClass = pyzo.parser.getFictiveClass(
                        className, self, handleSelf
                    )
                    handleSelf = False
                    if fictiveClass:
                        aco.addNames(fictiveClass.members)
                        classNames.extend(fictiveClass.supers)
                else:
                    nameForShell = className
                    break

        # If there's a shell, let it finish the autocompletion
        shell = pyzo.shells.getCurrentShell()
        if shell:
            aco.name = nameForShell  # might be the same or a base class
            shell.processAutoComp(aco)
        else:
            # Otherwise we finish it ourselves
            aco.finish()


if __name__ == "__main__":
    # Do some stubbing to run this module as a unit separate from pyzo
    # TODO: untangle pyzo from this module where possible
    class DummyParser:
        def parseThis(self, x):
            pass

    pyzo.parser = DummyParser()
    EditorContextMenu = QtWidgets.QMenu  # noqa
    app = QtWidgets.QApplication([])
    win = PyzoEditor(None)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+C"), win).activated.connect(win.copy)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+X"), win).activated.connect(win.cut)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+V"), win).activated.connect(win.paste)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Shift+V"), win).activated.connect(
        win.pasteAndSelect
    )
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Z"), win).activated.connect(win.undo)
    QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+Y"), win).activated.connect(win.redo)

    tmp = "foo(bar)\nfor bar in range(5):\n  print bar\n"
    tmp += "\nclass aap:\n  def monkey(self):\n    pass\n\n"
    win.setPlainText(tmp)
    win.show()
    app.exec_()



#
# Copyright © The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
import warnings


def introduce_renamed_methods_qheaderview(QHeaderView):

    _isClickable = QHeaderView.isClickable

    def sectionsClickable(self):
        """
        QHeaderView.sectionsClickable() -> bool
        """
        return _isClickable(self)

    QHeaderView.sectionsClickable = sectionsClickable

    def isClickable(self):
        warnings.warn(
            "isClickable is only available in Qt4. Use " "sectionsClickable instead.",
            stacklevel=2,
        )
        return _isClickable(self)

    QHeaderView.isClickable = isClickable

    _isMovable = QHeaderView.isMovable

    def sectionsMovable(self):
        """
        QHeaderView.sectionsMovable() -> bool
        """
        return _isMovable(self)

    QHeaderView.sectionsMovable = sectionsMovable

    def isMovable(self):
        warnings.warn(
            "isMovable is only available in Qt4. Use " "sectionsMovable instead.",
            stacklevel=2,
        )
        return _isMovable(self)

    QHeaderView.isMovable = isMovable

    _resizeMode = QHeaderView.resizeMode

    def sectionResizeMode(self, logicalIndex):
        """
        QHeaderView.sectionResizeMode(int) -> QHeaderView.ResizeMode
        """
        return _resizeMode(self, logicalIndex)

    QHeaderView.sectionResizeMode = sectionResizeMode

    def resizeMode(self, logicalIndex):
        warnings.warn(
            "resizeMode is only available in Qt4. Use " "sectionResizeMode instead.",
            stacklevel=2,
        )
        return _resizeMode(self, logicalIndex)

    QHeaderView.resizeMode = resizeMode

    _setClickable = QHeaderView.setClickable

    def setSectionsClickable(self, clickable):
        """
        QHeaderView.setSectionsClickable(bool)
        """
        return _setClickable(self, clickable)

    QHeaderView.setSectionsClickable = setSectionsClickable

    def setClickable(self, clickable):
        warnings.warn(
            "setClickable is only available in Qt4. Use "
            "setSectionsClickable instead.",
            stacklevel=2,
        )
        return _setClickable(self, clickable)

    QHeaderView.setClickable = setClickable

    _setMovable = QHeaderView.setMovable

    def setSectionsMovable(self, movable):
        """
        QHeaderView.setSectionsMovable(bool)
        """
        return _setMovable(self, movable)

    QHeaderView.setSectionsMovable = setSectionsMovable

    def setMovable(self, movable):
        warnings.warn(
            "setMovable is only available in Qt4. Use " "setSectionsMovable instead.",
            stacklevel=2,
        )
        return _setMovable(self, movable)

    QHeaderView.setMovable = setMovable

    _setResizeMode = QHeaderView.setResizeMode

    def setSectionResizeMode(self, *args):
        """
        QHeaderView.setSectionResizeMode(QHeaderView.ResizeMode)
        QHeaderView.setSectionResizeMode(int, QHeaderView.ResizeMode)
        """
        _setResizeMode(self, *args)

    QHeaderView.setSectionResizeMode = setSectionResizeMode

    def setResizeMode(self, *args):
        warnings.warn(
            "setResizeMode is only available in Qt4. Use "
            "setSectionResizeMode instead.",
            stacklevel=2,
        )
        _setResizeMode(self, *args)

    QHeaderView.setResizeMode = setResizeMode



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Yoton is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

# flake8: noqa

import sys
import time

import yoton
from yoton.misc import basestring, bytes, str
from yoton.misc import Property, getErrorMsg, UID
from yoton.misc import PackageQueue

from yoton.connection import Connection, TIMEOUT_MIN
from yoton.connection import STATUS_CLOSED, STATUS_WAITING, STATUS_HOSTING
from yoton.connection import STATUS_CONNECTED, STATUS_CLOSING


class ItcConnection(Connection):
    """ItcConnection(context, hostname, port, name='')

    Not implemented .

    The inter-thread-communication connection class implements a
    connection between two contexts that are in the same process.
    Two instances of this class are connected using a weak reference.
    In case one of the ends is cleaned up by the garbadge collector,
    the other end will close the connection.

    """

    pass
    # todo: implement me



# -*- coding: utf-8 -*-
# Copyright (C) 2013, the Pyzo development team
#
# Pyzo is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.

""" pyzo.util._locale
Module for locale stuff like language and translations.
"""

import os, sys, time

import pyzo
from pyzo.qt import QtCore, QtWidgets

QLocale = QtCore.QLocale

# Define supported languages. The key defines the name as shown to the
# user. The value is passed to create a Locale object. From the local
# object we obtain the name for the .tr file.
# Chinese:
LANGUAGES = {
    "English (US)": QLocale.C,
    # == (QLocale.English, QLocale.UnitedStates),
    #'English (UK)': (QLocale.English, QLocale.UnitedKingdom),
    "Dutch": QLocale.Dutch,
    "Spanish": QLocale.Spanish,
    "Catalan": QLocale.Catalan,
    "French": QLocale.French,
    "German": QLocale.German,
    "Russian": QLocale.Russian,  # not updated for 3.4
    "Polish": QLocale.Polish,
    "Portuguese": QLocale.Portuguese,
    "Portuguese (BR)": (QLocale.Portuguese, QLocale.Brazil),
    "Simplified Chinese": QLocale.Chinese,
    "Traditional Chinese": (
        QLocale.Chinese,
        QLocale.Taiwan,
    ),  # https://bugreports.qt.io/browse/QTBUG-1573
    # Languages for which the is a .tr file, but no translations available yet:
    # 'Slovak': QLocale.Slovak,
}


LANGUAGE_SYNONYMS = {
    None: "English (US)",
    "": "English (US)",
    "English": "English (US)",
    "ca_ES": "Catalan",
    "de_DE": "German",
    "es_ES": "Spanish",
    "fr_FR": "French",
    "nl_NL": "Dutch",
    "pl_PL": "Polish",
    "pt_BR": "Portuguese (BR)",
    "pt_PT": "Portuguese",
    "ru_RU": "Russian",
    "zh_CN": "Simplified Chinese",
    "zh_TW": "Traditional Chinese",
}


def getLocale(languageName):
    """getLocale(languageName)
    Get the QLocale object for the given language (as a string).
    """

    # Try System  Language if nothing defined
    if languageName == "":
        languageName = QLocale.system().name()

    # Apply synonyms
    languageName = LANGUAGE_SYNONYMS.get(languageName, languageName)

    # if no language applicable, get back to default
    if LANGUAGES.get(languageName, None) is None:
        languageName = LANGUAGE_SYNONYMS.get("", "")

    # Select language in qt terms
    qtLanguage = LANGUAGES.get(languageName, None)
    if qtLanguage is None:
        raise ValueError("Unknown language")

    # Return locale
    if isinstance(qtLanguage, tuple):
        return QLocale(*qtLanguage)
    else:
        return QLocale(qtLanguage)


def setLanguage(languageName):
    """setLanguage(languageName)
    Set the language for the app. Loads qt and pyzo translations.
    Returns the QLocale instance to pass to the main widget.
    """

    # Get locale
    locale = getLocale(languageName)

    # Get paths were language files are
    qtTransPath = str(
        QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.TranslationsPath)
    )
    pyzoTransPath = os.path.join(pyzo.pyzoDir, "resources", "translations")

    # Get possible names for language files
    # (because Qt's .tr files may not have the language component.)
    localeName1 = locale.name()
    localeName2 = localeName1.split("_")[0]

    # Uninstall translators
    if not hasattr(QtCore, "_translators"):
        QtCore._translators = []
    for trans in QtCore._translators:
        QtWidgets.QApplication.removeTranslator(trans)

    # The default language
    if localeName1 == "C":
        return locale

    # Set Qt translations
    # Note that the translator instances must be stored
    # Note that the load() method is very forgiving with the file name
    for what, where in [("qt", qtTransPath), ("pyzo", pyzoTransPath)]:
        trans = QtCore.QTranslator()
        # Try loading both names
        for localeName in [localeName1, localeName2]:
            success = trans.load(what + "_" + localeName + ".tr", where)
            if success:
                QtWidgets.QApplication.installTranslator(trans)
                QtCore._translators.append(trans)
                print("loading %s %s: ok" % (what, languageName))
                break
        else:
            print("loading %s %s: failed" % (what, languageName))

    # Done
    return locale


class Translation(str):
    """Derives from str class. The translate function returns an instance
    of this class and assigns extra atrributes:
      * original: the original text passed to the translation
      * tt: the tooltip text
      * key: the original text without tooltip (used by menus as a key)

    We adopt a simple system to include tooltip text in the same
    translation as the label text. By including ":::" in the text,
    the text after that identifier is considered the tooltip.
    The text returned by the translate function is always the
    string without tooltip, but the text object has an attribute
    "tt" that stores the tooltip text. In this way, if you do not
    use this feature or do not know about this feature, everything
    keeps working as expected.
    """

    pass


def _splitMainAndTt(s):
    if ":::" in s:
        parts = s.split(":::", 1)
        return parts[0].rstrip(), parts[1].lstrip()
    else:
        return s, ""


def translate(context, text, disambiguation=None):
    """translate(context, text, disambiguation=None)
    The translate function used throughout pyzo.
    """
    # Get translation and split tooltip
    newtext = QtCore.QCoreApplication.translate(context, text, disambiguation)
    s, tt = _splitMainAndTt(newtext)
    # Create translation object (string with extra attributes)
    translation = Translation(s)
    translation.original = text
    translation.tt = tt
    translation.key = _splitMainAndTt(text)[0].strip()
    return translation


## Development tools
import subprocess

LHELP = """
Language help - info for translaters

For translating, you will need a set of working Qt language tools:
pyside-lupdate, linguist, lrelease. On Windows, these should come
with your PySide installation. On (Ubuntu) Linux, you can install
these with 'sudo apt-get install pyside-tools qt4-dev-tools'.

You also need to run pyzo from source as checked out from the repo
(e.g. by running pyzolauncher.py).

To create a new language:
  * the file 'pyzo/util/locale.py' should be edited to add the language
    to the LANGUAGES dict
  * run 'linguist(your_lang)', this will raise an erro, but it will show
    the name of the .tr file
  * the file 'pyzo/pyzo.pro' should be edited to include the new .tr file
  * run 'lupdate()' to create the .tr file
  * run 'linguist(your_lang)' again to initialize the .tr file.

To update a language:
  * run 'lupdate()'
  * run 'linguist(your_lang)'
  * make all the translations and save
  * run lrelease() and restart pyzo to see translations
  * repeat if necessary

"""


def lhelp():
    """lhelp()
    Print help text on using the language tools.
    """
    print(LHELP)


def linguist(languageName):
    """linguist(languageName)
    Open linguist with the language file as specified by lang. The
    languageName can be one of the fields as visible in the language
    list in the menu. This function is intended for translators.
    """
    # Get locale
    locale = getLocale(languageName)

    # Get file to open
    fname = "pyzo_{}.tr".format(locale.name())
    filename = os.path.join(pyzo.pyzoDir, "resources", "translations", fname)
    if not os.path.isfile(filename):
        raise ValueError("Could not find {}".format(filename))

    # PyQt5 does not come with linguist anymore? Install PySide2 and check the
    # pyside2 package directory for the linguist exe ...

    # Get Command for linguist
    qtcore_mod_name = pyzo.QtCore.QObject.__module__
    qtcore_mod_path = sys.modules[qtcore_mod_name].__file__
    pysideDir = os.path.abspath(os.path.dirname(qtcore_mod_path))
    print(pysideDir)
    ISWIN = sys.platform.startswith("win")
    exe_ = "linguist" + ".exe" * ISWIN
    exe = os.path.join(pysideDir, exe_)
    if not os.path.isfile(exe):
        exe = exe_

    # Spawn process
    return subprocess.Popen([exe, filename])


def lupdate():
    """For developers. From pyzo.pro create the .tr files"""
    # Get file to open
    fname = "pyzo.pro"
    filename = os.path.realpath(os.path.join(pyzo.pyzoDir, "..", fname))
    if not os.path.isfile(filename):
        raise ValueError(
            "Could not find {}. This function must run from the source repo.".format(
                fname
            )
        )

    # Get Command for python lupdate
    pysideDir = os.path.abspath(os.path.dirname(pyzo.QtCore.__file__))
    ISWIN = sys.platform.startswith("win")
    exe_ = "pylupdate" + pyzo.QtCore.__version__[0] + ".exe" * ISWIN
    exe = os.path.join(pysideDir, exe_)
    if not os.path.isfile(exe):
        exe = exe_

    # Spawn process
    cmd = [exe, "-noobsolete", "-verbose", filename]
    p = subprocess.Popen(
        cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    while p.poll() is None:
        time.sleep(0.1)
    output = p.stdout.read().decode("utf-8")
    if p.returncode:
        raise RuntimeError("lupdate failed (%i): %s" % (p.returncode, output))
    else:
        print(output)


def lrelease():
    """For developers. From pyzo.pro and the .tr files, create the .qm files."""
    # Get file to open
    fname = "pyzo.pro"
    filename = os.path.realpath(os.path.join(pyzo.pyzoDir, "..", fname))
    if not os.path.isfile(filename):
        raise ValueError(
            "Could not find {}. This function must run from the source repo.".format(
                fname
            )
        )

    # Get Command for lrelease
    pysideDir = os.path.abspath(os.path.dirname(pyzo.QtCore.__file__))
    ISWIN = sys.platform.startswith("win")
    exe_ = "lrelease" + ".exe" * ISWIN
    exe = os.path.join(pysideDir, exe_)
    if not os.path.isfile(exe):
        exe = exe_

    # Spawn process
    cmd = [exe, filename]
    p = subprocess.Popen(
        cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    while p.poll() is None:
        time.sleep(0.1)
    output = p.stdout.read().decode("utf-8")
    if p.returncode:
        raise RuntimeError("lrelease failed (%i): %s" % (p.returncode, output))
    else:
        print(output)


if __name__ == "__main__":
    # Print names of translator files

    print("Language data files:")
    for key in LANGUAGES:
        s = "{}: {}".format(key, getLocale(key).name() + ".tr")
        print(s)



# -*- coding: utf-8 -*-
# flake8: noqa
# Copyright (C) 2013, the Pyzo development team
#
# Yoton is distributed under the terms of the 2-Clause BSD License.
# The full license can be found in 'license.txt'.


"""
Yoton is a Python package that provides a simple interface
to communicate between two or more processes.

Yoton is ...
  * lightweight
  * written in pure Python
  * without dependencies (except Python)
  * available on Python version >= 2.4, including Python 3
  * cross-platform
  * pretty fast

"""

# Import stuff from misc and events
from yoton.misc import UID, str, bytes
from yoton.events import Signal, Timer, app

# Inject app function in yoton namespace for convenience
call_later = app.call_later
process_events = app.process_events
start_event_loop = app.start_event_loop
stop_event_loop = app.stop_event_loop
embed_event_loop = app.embed_event_loop

# Import more
from yoton.core import Package
from yoton.connection import Connection, ConnectionCollection
from yoton.connection_tcp import TcpConnection
from yoton.context import Context
from yoton.clientserver import RequestServer, do_request
from yoton.channels import *


# Set yoton version
__version__ = "2.2"


# Define convenience class
class SimpleSocket(Context):
    """SimpleSocket()

    A simple socket has an API similar to a BSD socket. This socket
    sends whole text messages from one end to the other.

    This class subclasses the Yoton.Context class, which makes setting
    up this socket very easy.

    Example
    -------
    # One end
    s = SimpleSocket()
    s.bind('localhost:test')
    s.send("Hi")

    # Other end
    s = SimpleSocket()
    s.connect('localhost:test')
    print(s.recv())

    """

    def __init__(self, verbose=False):
        Context.__init__(self, verbose)

        # Create channels
        self._cs = PubChannel(self, "text")
        self._cr = SubChannel(self, "text")

    def send(self, s):
        """send(message)

        Send a text message. The message is queued and send
        over the socket by the IO-thread.

        """
        self._cs.send(s)

    def recv(self, block=None):
        """recv(block=None):

        Read a text from the channel. What was send as one message is
        always received as one message.

        If the channel is closed and all messages are read, returns ''.

        """
        return self._cr.recv(block)
